      PROGRAM LVG
c
c    LVG calculates a 1-D, 2-D, or 3-D grid of lvg radiative transfer
c    models as specified by the input parameters.
c
c  History:
c    lgm  pre-his   general lvg program exists outside of MIRIAD. Dinosars
c                      roam the earth.
c    lgm  history   lvg program minimally converted to run under MIRIAD. 
c                      Mammals rise.
c    lgm  30jun92   slow movement toward full-filling MIRIAD standards.
c    lgm  02jul92   enhansed to do select transitions for output and
c                      run 3-D models.
c    lgm  13jul92   minor addition to format of output, corrected floating
c                      point error by changing tback to tbk, added check
c                      for existence of output file.
c    lgm  01jun94   added logclose to end of main
c    lgm  06aug97   updated log writes to be compatable with next miriad subs
c    pjt  02jul02   fortran (ANSI) fix statement for linux
c
c------ Inline doc (retrieved with doc to a .doc file) ----------------c
c
c= LVG - Calculates LVG radiative transfer models.
c& lgm
c: models
c+       
c    LVG calculates a 1-D, 2-D, or 3-D grid of lvg radiative transfer
c    models as specified by the input parameters. When a range of
c    values is specified for a parameter (density, temperature, fractional
c    abundance, column density), models are logrithmically spaced over the
c    interval.
c@ mol
c   Name of molecule. Known molecules are CO, CS, etc. No Default.
c@ iso
c   Isotope of molecule if other than main isotopic species. Default
c   is main isotopic species.
c@ den
c   Density range to be covered by the model grid. Input may be one, two,
c   or three numbers. The first number indicates the number of density
c   models to be calculated; the second and third numbers specify the density 
c   of the first and last models in particles/cm3. Models are logrithmically 
c   spaced. If only two numbers are given, the density for all models is set 
c   to the second number. If only one number is given, the default density 
c   is assumed. 
c   Example: den = 5,1.0e4,1.0e6 -- 5 models with densities from 1.0e4 to 1.0e6.
c   Defaults = 1,10000.
c@ xmol
c   Range of fractional abundance of molecule to be used in calculation. 
c   Input may be one, two, or three numbers as above. 
c   WARNING -- xmol and cd can not be non-zero simultaneously. 
c   Default = 0,1.0e-8
c@ cd
c   Range in molecular column density to be used in calculation. Input
c   may be one, two, or three numbers; see description of density for more 
c   format information. 
c   WARNING -- xmol and cd can not be non-zero simultaneously. 
c   Default = 1,1.0e12.
c@ tk 
c   Range of kinetic temperature to be used in calculating models. Input
c   may be one, two, or three numbers; see description of density for more 
c   format information. 
c   Default = 1,20
c@ dv
c   Velocity FWHM of line in km/sec. CD will be divided by this value
c   for calculating the radiative excitation.
c@ iter
c   Maximum number of iterations allowed for each model. Usually the defaults 
c   is fine but models with high opacity and low density may require more 
c   iterations.
c   Default = 25.
c@ tbkg
c   Blackbody temperature for background radiation field in all transitions. 
c   Normally set to the cosmic background temperature.
c   Default = 2.78.
c@ pops
c   Data file name for level populations. 
c   Two or three of den, xmol, cd, or tk must be specified with more than
c   1 model along that axis. If no name is given, no file will be written. 
c   Default: '  '
c@ trads
c   Data file name for line radiation temperatures.
c   Two of den, xmol, cd, or tk must be specified with more than
c   1 model along that axis. If no name is given, no file will be written. 
c   Default: '  '
c@ trans
c   List of frequencies (in GHz) of transitions to be included in 'trads' file. 
c   If none are specified, 20 transitions are selected. If there are more
c   than 20 possible transitions, the first 20 with frequencies between 70 GHz
c   and 500 GHz are selected. No more than 20 transitions may be written.
c   Default: program selects 20 transitions.
c@ print
c   Parameter which control level of printed output.
c      0 = minimum output which is useful when running grids
c      1 = normal output level
c      2 = detailed output level - used for debugging and error checking
c@ log
c   Name of log file were full printed output for human consumption
c   is written. 
c
c------------------------------------------------------------------------
c
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      INCLUDE 'cubes.h'
      CHARACTER*10 molcul,mold,isotop
      CHARACTER logfile*50,popname*50,tradname*50,cubename*50
      CHARACTER text*80,ddash*80,version*25
      REAL den1,den2,xmol1,xmol2,cd1,cd2,tk1,tk2
      INTEGER i,j,ix,itk,nden,nx,ncd,ntk,iprint,naxis,inam,imol
      REAL ddel,cdel,xdel,tdel,deni,xmoli,tki,trans(MAXTRAN)
      REAL pops(MAXLEV)
      INTEGER tout,iden
      LOGICAL more,savepop,savelin,there,keyprsnt
      INCLUDE 'radinit.h'
      DATA mold / 'OLDNAME' /
      DATA more / .true. /
      DATA version / ' Version July 2, 2002 ' /
      v0=3.30E-14
      DDASH(1:30)  = '=============================='
      DDASH(31:60) = '=============================='
      DDASH(61:80) = '===================='

C------------------------------------------------------------------
C WARNING -- THIS is JUST a temp fix
      do 5 i=1,maxks
         jmax(i) = -1
    5 continue
C-----------------------------------------------------------
      CALL output('LVG:'//version)
C--------------------------------------------------------
C   Get input using key routines
C
      CALL KEYINI
      CALL KEYA('mol',molcul,' ')
      if(molcul .eq. ' ') call bug('f','Molecule must be specified')
      CALL KEYA('iso',isotop,molcul)
C
C   Input and check densities
C
      CALL KEYI('den',nden,1)
      CALL KEYR('den',den1,1.0e4)
      CALL KEYR('den',den2,0.0)
      if(nden .lt. 0)   nden = 1
      if(nden .gt. MAXMOD) then
         call bug('w',' Number of density models reset to max allowed')
         nden = MAXMOD
      endif
      if(den1 .lt. 0.0) then
         call bug('w',' Density <= 0.0; Reset to default')
         den1 = 1.0e4
      endif
      if(den2 .le. 0.0) nden = 1
C
C   Input and check fractional abundances
C
      CALL KEYI('xmol',nx,0)
      CALL KEYR('xmol',xmol1,1.0e-8)
      CALL KEYR('xmol',xmol2,0.0)
      if(nx .lt. 0) nx = 0
C
C   Input and check column densities
c
      CALL KEYI('cd',ncd,1)
      CALL KEYR('cd',cd1,1.0e12)
      CALL KEYR('cd',cd2,0.0)
      if(ncd .lt. 0) ncd = 0
      if(ncd .gt. MAXMOD) then
         call bug('w',' Number of col den models reset to max allowed')
         ncd = MAXMOD
      endif
      if(ncd .gt. 0 .and. nx .gt. 0) 
     * call bug('f','First values of xmol and cd can not both be >0')
C
C   Input and check temperatures
C
      CALL KEYI('tk',ntk,1)
      CALL KEYR('tk',tk1,20.0)
      CALL KEYR('tk',tk2,0.0)
      if(ntk .gt. MAXMOD) then
         call bug('w',' Number of temp models reset to max allowed')
         ntk = MAXMOD
      endif
C
C   Input velocity width
C
      CALL KEYR('dv',dopv,1.0)
      if(dopv .lt. 0.0) then
	 call bug('w','dv must be positive; absolute value taken')
         dopv = -dopv
      endif
C
C   Input and check iter and tbkg
C
      CALL KEYI('iter',nit,50)
      if(nit .gt. 1000 .or. nit .le. 0) then
         call bug('w','ITER out of range 1 to 1000; Reset to default')
         nit = 50
      endif
      CALL KEYR('tbkg',tbk,2.78)
      if(tbk .gt. 1.0e5 .or. tbk .lt. 0.0) then
         call bug('w','TBGK out of range 0 to 1e5; Reset to default')
         tbk = 2.78
      endif
C
C   Get name for output cubes containing population and line temperatures
C
      savepop = .false.
      savelin = .false.
      cubename = '     '
      CALL KEYA('pops',popname,' ')
      if(popname(1:1) .ne. ' ') then
         inquire(file=popname,exist=there)
         if(there) call bug('f',' pops output file already exists')
	 cubename = popname
	 savepop  = .true.
	 savelin  = .false.
      endif
      CALL KEYA('trads',tradname,' ')
      if(tradname(1:1) .ne. ' ') then
	 inquire(file=tradname,exist=there)
	 if(there) call bug('f',' trads output file already exists')
	 cubename = tradname 
	 savepop  = .false.
	 savelin  = .true.
      endif
      if(popname(1:1) .ne. ' ' .and. tradname(1:1) .ne. ' ')
     1	call bug('f','Both popname and tradname cannot be specified')
C
C   Get transitions freq selecting which lines to write to trad file
C
      ntrans = 0
      dowhile(keyprsnt('trans'))
	 ntrans = ntrans + 1
  	 call KEYR('trans',trans(ntrans),0.0)
      enddo
C
C   Input print level specification
C
      CALL KEYI('print',iprint,1)
C
C   Get specification of log file and grid file
C
      CALL KEYA('log',logfile,' ')
C
      CALL KEYFIN
C
C   End of input from key routines
C------------------------------------------------------------
C
C   Figure out information for axis of models
C
      nsize(3) = 1
      nsize(4) = 1
      naxis = 1
      if(nden .gt. 1) then
	 nsize(naxis) = nden
	 ctyp(naxis)  = 'density'
	 crval(naxis) = alog10(den1)
	 ddel         = (alog10(den2)-alog10(den1))/float(nden-1)
	 crdel(naxis) = ddel
	 naxis        = naxis + 1
      endif
      if(ncd .gt. 1) then
	 nsize(naxis) = ncd
	 ctyp(naxis)  = 'col-den'
         crval(naxis) = alog10(cd1)
         cdel         = (alog10(cd2)-alog10(cd1))/float(ncd-1)
	 crdel(naxis) = cdel
	 naxis        = naxis + 1
      elseif(nx .gt. 1) then
         nsize(naxis) = nx
         ctyp(naxis)  = 'fra-abun'   
	 crval(naxis) = alog10(xmol1)
         xdel         = (alog10(xmol2)-alog10(xmol1))/float(nx-1)
	 crdel(naxis) = xdel
         naxis        = naxis + 1
      endif
      if(ntk .gt. 1) then
	 nsize(naxis) = ntk
	 ctyp(naxis)  = 'kin-temp'
	 crval(naxis) = alog10(tk1)
         tdel         = (alog10(tk2)-alog10(tk1))/float(ntk-1)
	 crdel(naxis) = tdel
	 naxis        = naxis + 1
      endif
      if(cubename(1:1) .ne. ' ') then
	 if(naxis .lt. 3) call bug('f',
     1		'Two or more of den,(cd or xmol) and, tk must be >1')
	 crval(naxis) = 1.0
	 crdel(naxis) = 1.0
      endif
C
C   Open log file of specified
C 
      call logopen(logfile,' ')
C-----------------------------------------------------------
C 
      DO 15 I=1,NDEN
         RNH2(I)=den1*(10.0**(FLOAT(I-1)*DDEL))
   15 CONTINUE 
      IF(NX.GE.1) THEN
         DO 20 I=1,NX
            RNX(I)=xmol1*(10.0**(FLOAT(I-1)*XDEL))
   20    CONTINUE
      ELSE
         DO 25 I=1,NCD
            RCD(I)=cd1*(10.0**(FLOAT(I-1)*CDEL))
   25    CONTINUE
         NX=NCD
      ENDIF 
      DO 30 I=1,NTK
         RTK(I)=tk1*(10.0**(FLOAT(I-1)*TDEL))
   30 CONTINUE
C------------------------------------------------------
C  This section writes out the model parameters to the
C  chosen output location
C
      WRITE(text,1998)
      call LogWrite(text,more)
      call LogWrite(' ',more)
      WRITE(text,1999) ISOTOP
      call LogWrite(text,more)
      IF(NDEN.GT.1) THEN
         WRITE(text,2000) RNH2(1),NDEN,RNH2(NDEN)
         call LogWrite(text,more)
         WRITE(text,2002)  DDEL
      ELSE
         WRITE(text,2001) RNH2(1)
      ENDIF
      call LogWrite(text,more)
      IF(NCD.GT.0) THEN
         IF(NCD.GT.1) THEN
            WRITE(text,2005) ISOTOP,RCD(1),NCD,RCD(NCD)
            call LogWrite(text,more)
            WRITE(text,2004) CDEL
         ELSE
            WRITE(text,2006) ISOTOP,RCD(1)
         ENDIF
         call LogWrite(text,more)
      ELSE
         IF(NX.GT.1) THEN
            WRITE(text,2010) ISOTOP,RNX(1),NX,RNX(NX)
            call LogWrite(text,more)
            WRITE(text,2009) XDEL
         ELSE
            WRITE(text,2011) ISOTOP,RNX(1)
         ENDIF
         call LogWrite(text,more)
      ENDIF
      IF(NTK.GT.1) THEN
         WRITE(text,2020) RTK(1),NTK,RTK(NTK)
         call LogWrite(text,more)
         WRITE(text,2019) TDEL
      ELSE
         WRITE(text,2021) RTK(1)
      ENDIF
      call LogWrite(text,more)
      WRITE(text,2030) dopv
      call LogWrite(text,more)
 1998 FORMAT(20X,'SPHERICALLY SYMMETRIC LVG MODELS')
 1999 FORMAT(2X,16HINPUT QUANTITIES,6X,'Molecule name = ',A10)
 2000 FORMAT(5X,'Density at model 1 = ',1PE9.2,2X,'at model ',I2,
     1        ' = ',1PE9.2)
 2002 FORMAT(15X,'steps size = ',0PF5.2,' in the log')
 2001 FORMAT(5X,'Density fixed at ',1PE9.2,' for all models')
 2005 FORMAT(5X,'N(',A10,') at model 1 = ',1PE9.2,2X,'at model ',I2,
     1        ' = ',1PE9.2)
 2004 FORMAT(15X,'step size = ',0PF5.2,' in the log')
 2006 FORMAT(5X,'N(',A10,') fixed at ',1PE9.2,' for all models')
 2010 FORMAT(5X,'X(',A10,') at model 1 = ',1PE9.2,2X,'at model ',I2,
     1        ' = ',1PE9.2)
 2009 FORMAT(15X,'step size = ',0PF5.2,' in the log')
 2011 FORMAT(5X,'X(',A10,') fixed at ',1PE9.2,' for all models')
 2020 FORMAT(5X,'T kin at model 1 = ',F5.1,2X,'at model ',I2,' = ',
     1        F5.1)
 2019 FORMAT(15X,'step size = ',F5.1,' in the log')
 2021 FORMAT(5X,'T kin fixed at ',F5.1,' for all models')
 2030 FORMAT(5X,'Velocity gradient fixed at',f6.1,' km/s/pc')
C 
C End of loop for entering and translating model parameters
C------------------------------------------------------------ 
C  This area of code reads in the collision rates and, if necessary
C  reads in the Einstein coefficients. For linear molecules
C  IMOL=1; for asymmetric molecules, IMOL=-1. If isotopic
C  corrections are necessary, they are done here.
C
      call LogWrite(' ',more)
      CALL MOLREAD(MOLCUL,INAM,IMOL)
      IF(INAM.EQ.0) CALL bug('f','Molecule not found')
      IF(MOLD.NE.MOLCUL) THEN
	 CALL ISOID(ISOTOP,IMOL)
         DYOLD = DY
         CALL COLL1(IPRINT,ISOTOP,IMOL)
         MOLD=MOLCUL
      ENDIF
 4000 FORMAT(15X,E9.4,22X,E8.2,18X,E8.2)
 4010 FORMAT(18X,E11.5,7X,E10.4,7X,E10.4)
 4020 FORMAT(20X,F5.2)
      call LogWrite(' ',more)
      IF(IMOL.EQ.1) THEN
         WRITE(text,2040) BROT,DE 
      ELSEIF(IMOL .eq. -1) THEN
         WRITE(text,2041) A0,B0,C0
      ELSE
	 WRITE(text,2042) B0,C0,DJ
      ENDIF
      call LogWrite(text,more)
      WRITE(text,2050) DY 
      call LogWrite(text,more)
 2040 FORMAT(5X,'Rotational Constant = ',f10.5,' GHz',3X, 
     1       'Distortion Constant = ',1PE10.3,' GHz') 
 2041 FORMAT(5X,'A0 = ',1PE10.3,2X,'B0 = ',E10.3,2X,'C0 = ',E10.3)
 2042 FORMAT(5X,'B0 = ',1Pe10.3,2x,'C0 = ',E10.3,2x,'DJ = ',E10.3)
 2050 FORMAT(5X,'Dipole moment = ',F4.2,' Dbye') 
C
C   Fill array of possible transitions with background temperature
C
      DO i=1,nlev
      DO J=1,nlev
         BBK(i,j) = tbk
      ENDDO
      ENDDO
      CALL LogWrite(ddash,more)
c
c   Fill out array of logicals for printing out MIRIAD data file
c
      CALL seltrans(trans,ntrans)
c
C-------------------------------------------------------
C These are the actual do loops for calculating the multiple
C models
C  
      DO 400 ITK=1,NTK
         TKI=RTK(ITK)
	 CALL COLL2(TKI)
         DO 300 IX=1,NX
            DO 200 IDEN=1,NDEN
               DENI=RNH2(IDEN)
               IF(NCD.LT.1) THEN
                  XMOLI=RNX(IX)
               ELSE
                  XMOLI=RCD(IX)*V0/(DENI*1.0645E+5)
               ENDIF
               CALL ITER(ISOTOP,DENI,XMOLI,TKI,IPRINT)
               if(savelin) then
 	         call savecube(naxis,iden,ix,itk,trad,ifre)
	       endif
               if(savepop) then
                 do i=1,nlev
                   pops(i) = xn(i)
                 enddo
                 call savecube(naxis,iden,ix,itk,pops,nlev)
               endif
  200       CONTINUE 
  300    CONTINUE
  400 CONTINUE
C-------------------------------------------------------
C
C   Write out data cubes and close files, if opened
C
      if(cubename .ne. ' ') then
         if(savelin) then
            ctyp(naxis)  = 'lines'
            nsize(naxis) = ifre
         else
            ctyp(naxis)  = 'pops'
            nsize(naxis) = ifre
         endif
         call xyopen(tout,cubename,'new',naxis,nsize)
         call wrhead(tout,naxis,molcul)
         call putimage(tout,naxis)
	 call hisopen(tout,'write')
	 call hiswrite(tout,'LVG:'//version)
         call hisinput(tout,'lvg')
         if(savelin) then
	    call hiswrite(tout,'LVG: Cube of line radiation temps.')
	    call hiswrite(tout,'LVG: Plane-Transition Correspondence')
	    do i=1,ifre
	       write(text,'(''LVG: Plane # '',i3,'' Freq: '',f8.2)')
     1			i,trans(i)
	       call LogWrite(text,more)
	       call hiswrite(tout,text)
	    enddo
	 else
	    call hiswrite(tout,'LVG: Cube of level pops.')
	    call hiswrite(tout,'LVG: Plane-Level Correspondence')
	 endif
	 call hisclose(tout)
         call xyclose(tout)
      endif
      call logclose
C------------------------------------------------------------------
 3000 FORMAT(4(I2,2X))
 3005 FORMAT(12I6)
 3010 FORMAT(5(1PE10.3))
 3020 FORMAT(10(F5.1,1X)) 
      CLOSE(UNIT=11)
      CLOSE(UNIT=12)
      STOP 
      END 
C********************************************************** 
C**********END OF MAIN************************************ 
C********************************************************** 
	subroutine wrhead(lout,naxis,molcul)
C
C  writes header for image cube
C
	include 'rad_size.h'
	include 'cubes.h'
	integer lout,naxis
	character molcul*10
C
	call wrhdi(lout,'naxis',naxis)
	call wrhdi(lout,'naxis1',nsize(1))
        call wrhdi(lout,'naxis2',nsize(2))
        call wrhdi(lout,'naxis3',nsize(3))
	call wrhda(lout,'ctype1',ctyp(1))
        call wrhda(lout,'ctype2',ctyp(2)) 
        call wrhda(lout,'ctype3',ctyp(3)) 
        call wrhdd(lout,'crpix1',1.0d0)
        call wrhdd(lout,'crpix2',1.0d0)
        call wrhdd(lout,'crpix3',1.0d0)
	call wrhdd(lout,'cdelt1',dble(crdel(1)))
        call wrhdd(lout,'cdelt2',dble(crdel(2))) 
        call wrhdd(lout,'cdelt3',dble(crdel(3))) 
	call wrhdd(lout,'crval1',dble(crval(1)))
        call wrhdd(lout,'crval2',dble(crval(2))) 
        call wrhdd(lout,'crval3',dble(crval(3))) 
	call wrhda(lout,'object',molcul)
	if(naxis .eq. 4) then
	   call wrhdi(lout,'naxis4',nsize(4))
	   call wrhda(lout,'ctype4',ctyp(4))
	   call wrhdd(lout,'crpix4',1.0d0)
	   call wrhdd(lout,'cdelt4',dble(crdel(4)))
	   call wrhdd(lout,'crval4',dble(crval(4)))
	endif
	return
	end
C
	subroutine putimage(lout,naxis)
C
C  writes data cube out to image file
C
	include 'rad_size.h'
	include 'cubes.h'
	integer lout,i,j,k,l,iax(2),naxis
	real row(MAXMOD)
C
	if(naxis .eq. 4) then
	   do l=1,nsize(4)
	   do k=1,nsize(3)
	      iax(1) = k
	      iax(2) = l
	      call xysetpl(lout,2,iax)
	      do j=1,nsize(2)
	         do i=1,nsize(1)
	    	    row(i) = cube(i,j,k,l)
	         enddo
	         call xywrite(lout,j,row)
	      enddo
	   enddo
	   enddo
	else
	   do k=1,nsize(3)
              call xysetpl(lout,1,k)
              do j=1,nsize(2)
                 do i=1,nsize(1)
                    row(i) = cube(i,j,1,k)
                 enddo
                 call xywrite(lout,j,row)
              enddo
           enddo
	endif
C
	return
	end
C
	subroutine savecube(naxis,iden,ix,itk,values,nvals)
C
C  write the level populations into an array for later writing to a
C  data file
C
	include 'rad_size.h'
	include 'cubes.h'
	real values(MAXTRAN)
	integer iden,ix,itk,nvals,i,ii(3),naxis
	ii(3) = 1
        if (naxis.lt.3) then
           call bug('w','SAVECUBE: NAXIS too small')
           return
        endif
	do i=1,naxis-1
	  if(ctyp(i)(1:1) .eq. 'd') ii(i) = iden
	  if(ctyp(i)(1:1) .eq. 'c') ii(i) = ix
	  if(ctyp(i)(1:1) .eq. 'f') ii(i) = ix
	  if(ctyp(i)(1:1) .eq. 'k') ii(i) = itk
	enddo
        do i=1,nvals
           cube(ii(1),ii(2),ii(3),i) = values(i)
        enddo 
        return
        end
	
      SUBROUTINE ITER(ISOTOP,DENI,XMOLI,TKI,IPRINT) 
C------------------------------------------------------ 
C     ITERATES SOLUTION OF STAT. EQUIL. EQUATION 
C------------------------------------------------------ 
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      INTEGER i,j,k,jk,itt,iprint
      REAL DENI,XMOLI,TKI,told,bpl
      REAL S(MAXLEV,MAXLEV),XJ(MAXLEV,MAXLEV)
      DOUBLE PRECISION EV(MAXLEV),EVR(MAXLEV),FAC,F1,TAU,DIF,SUM
      DOUBLE PRECISION STEQ(MAXLEV,MAXLEV),BETA(MAXLEV,MAXLEV)
      DOUBLE PRECISION STIMCOR
      CHARACTER*10 ISOTOP 
      TOLD=TKI 
C==========================================================
C   Only do the following stuff the first time through
C
      IF(XN(1) .LE. 0.0) THEN
         DO 20 I=1,NLEV
         DO 20 J=1,NLEV
   20    XJ(I,J)=0.0
C---------------------------------------------------------
C  Calculate a thermal distribution for the fractional
C  populations of the first model.
C
         CALL XLTE(XN,TKI)
C
C   End of stuff which is only done the first time through
C==========================================================
      ENDIF
      F1=2.0*AH*1.0E+18/(AC*AC)
      FAC=1.0D-9*AH*AC*XMOLI*DENI/(4.0*PI*V0*dopv)
      ITT=0 
  250 CONTINUE 
C-------------------------------------------------------
C     Compute the source function for each transition
C
      DO 300 J=1,NLEV
      DO 300 K=1,NLEV
      IF(A(J,K).LE.0.0) GO TO 300 
	 STIMCOR = (XN(K)*DEGEN(J)/(XN(J)*DEGEN(K)))-1.0D0
         IF(abs(stimcor) .gt. 1.0e-7) then
            S(J,K) = F1*(FREQ(J,K)**3)/STIMCOR
	 else
	    S(J,K) = F1*(FREQ(J,K)**3)
	 endif
  300 CONTINUE 
C--------------------------------------------------------
C  Compute the escape probability
C
      DO 320 J=1,NLEV
      DO 320 K=1,NLEV
         IF(A(J,K).LE.0.0) GO TO 320 
         TAU=FAC*(XN(K)*BA(K,J)-XN(J)*BE(J,K))
         IF(TAU.GT.50.0) THEN
            BETA(J,K) =1.0/TAU
         ELSE IF(TAU .LT. 1.0e-3) THEN
            BETA(J,K) = 1.0D0-0.5D0*TAU+TAU*TAU/6.0D0
         ELSE
            BETA(J,K) = (1.0D0-DEXP(-TAU))/TAU
         ENDIF
  320 CONTINUE 
C-----------------------------------------------------------
C     Compute the mean intensity to be used for calculating
C     the radiative excitation.
C
      DO 340 J=1,NLEV
      DO 340 K=1,NLEV 
         IF(A(J,K).LE.0.0) GO TO 340
         BPL=F1*(FREQ(J,K)**3)/(EXP(AH*FREQ(J,K)/ 
     1       (AK*BBK(J,K)))-1.0) 
         STIMCOR = ((XN(K)*DEGEN(J))/(XN(J)*DEGEN(K)))-1.0D0
         IF(abs(stimcor) .gt. 1.0e-7) then
            XJ(J,K)=(1.0D0-BETA(J,K))*S(J,K)+BETA(J,K)*BPL 
	 else
	    XJ(J,K)=0.5*FAC*BA(K,J)*XN(J)*DEGEN(K)*S(J,K)/DEGEN(J) +
     1				BETA(J,K)*BPL
	 endif
  340 CONTINUE 
C---------------------------------------------------
C     Set up (STATEQ) and then solve (SOLVE) 
C     the Statistical Equilibrium Equation. 
C
      CALL STATEQ(DENI,XJ,STEQ,MAXLEV) 
      CALL SOLV(STEQ,XN,NLEV,EV) 
C----------------------------------------------------
C  Fix up the populations that are too small to be
C  accurate due to machine precision.
C
      DO 400 I=2,NLEV 
         IF(EV(I) .GT. 1.0E-13) GO TO 400 
         EV(I)=EV(I-1)*DEGEN(I)*EXP(-AH*(ENER(I)-ENER(I-1))/
     1         (AK*BBK(I,I)))/DEGEN(I-1) 
  400 CONTINUE 
C -----------------------------------------------------------------
C    Renormalize the fractional populations to insure that the
C    sum is 1.0 to machine precision.
C
      SUM=0.0 
      DO 430 I=1,NLEV
         EVR(I)=XN(I) 
         SUM=SUM+EV(I) 
  430 CONTINUE 
      DO 440 I=1,NLEV 
         XN(I)=EV(I)/SUM 
  440 CONTINUE 
C-----------------------------------------------------------
C   Check for convergence of populations in levels with
C   significant populations (ignore the last level because
C   it converges very slowly). If all significant levels are
C   not converged, go back to the start of the loop.
C 
      ITT=ITT+1 
      DO 470 JK=1,NLEV-1
         IF(XN(JK).LT.1.0E-11) GO TO 470 
         DIF=DABS(EVR(JK)-XN(JK)) 
         IF(DIF.LT.1.0E-13) GO TO 470 
         IF(XN(JK)/DIF .LT. 500D0) GO TO 475 
  470 CONTINUE 
      GO TO 480 
  475 IF(ITT .LT. NIT) GO TO 250 
  480 CONTINUE 
      CALL PRTOUT(ISOTOP,DENI,XMOLI,TKI,BETA,ITT,IPRINT)
      RETURN 
C*********************************************************** 
C*************End of ITER*********************************** 
C*********************************************************** 
      END 
      SUBROUTINE PRTOUT(ISOTOP,DENI,XMOLI,TKI,BETA,IT,IPRINT) 
C------------------------------------------------------------- 
C     Prints out the results of the S.E. calculations 
C------------------------------------------------------------- 
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      INTEGER i,j,j1,kl,ku,kk,it,iprint,nn
      REAL deni,xmoli,cden,tki,xnu,bbck,btex,btr,f1,f2,tex,tr
      CHARACTER*10 ISOTOP 
      CHARACTER TEXT*80
      LOGICAL MORE
      DOUBLE PRECISION BETA(MAXLEV,MAXLEV),TAU 
      DATA MORE /.true. /

      CDEN=DENI*XMOLI*1.0645E+5/V0
      call LogWrite(' ',more)
      WRITE(text,2010) ISOTOP,IT 
      call LogWrite(text,more)
      call LogWrite(' ',more)
      WRITE(text,2020) DENI,ISOTOP,XMOLI,CDEN
      call LogWrite(text,more)
      WRITE(text,2021) TKI
      call LogWrite(text,more)
      IF(IPRINT.GE.1) THEN
         NN=(NLEV/4)+1
         DO 10 I=1,NN 
            KL=(I-1)*4+1
            IF(KL.GT.NLEV) GO TO 10
            KU=KL+3
            IF(KU.GT.NLEV) KU=NLEV 
            WRITE(TEXT,2030) (KK-1,XN(KK),KK=KL,KU) 
            call LogWrite(text,more)
   10    CONTINUE
      ENDIF 
      WRITE(text,2050) 
      call LogWrite(text,more)
      F1=2.0E+18*AH/(AC*AC) 
      F2=1.0E-18*AC*AC/(2.0*AK) 
      FAC=1.0D-9*AH*AC*XMOLI*DENI/(4.0*PI*V0*dopv)
C------------------------------------------------------------
C Compute the line radiation temperature and print out
C   the interesting quantities for each transition. 
C
      ifre = 0
      DO 20 I=1,NLEV
      DO 20 J=1,NLEV 
         IF(A(I,J).GT.0.0) then
            CALL TAUBETA(BETA(I,J),TAU) 
            XNU  = FREQ(I,J) 
            BBCK = F1*(XNU**3)/(EXP(AH*XNU/(AK*BBK(I,J)))-1.0) 
	    STIMCOR = XN(I)*DEGEN(J)/(XN(J)*DEGEN(I)) - 1.0
	    if(abs(stimcor) .gt. 1.0e-7) then
               TEX  = -AH*XNU/(AK*DLOG(XN(I)*DEGEN(J)/(XN(J)*DEGEN(I))))
               BTEX = F1*(XNU**3)/(EXP(AH*XNU/(AK*TEX))-1.0) 
               IF(TAU.LT.20.0) THEN
                  BTR = (BTEX-BBCK)*(1.0D0-DEXP(-TAU)) 
               ELSE
                  BTR = (BTEX-BBCK)
               ENDIF
	    else
	       TEX = -AH*XNU/(AK*DLOG(1.0d0+1.0d-7))
	       BTR = F1*(XNU**3)*XN(I)*DEGEN(J)*FAC*BA(J,I)/DEGEN(I)
     1			  - BBCK*TAU
            endif
            TR = BTR*F2/(XNU*XNU) 
            J1 = J-1 
            WRITE(text,2060) J,J1,BETA(I,J),TEX,TAU,TR,XNU 
            call LogWrite(text,more)
	    if(wrtrans(i,j)) then
               ifre         = ifre + 1
	       trad(ifre)   = tr
	    endif
	 endif
   20 CONTINUE
 1000 FORMAT(F4.1)
 2010 FORMAT(15X,'LVG MODEL FOR ',A10,' converged in ',I3,
     1      ' iterations') 
 2020 FORMAT(5X,'Density = ',1PE9.2,' X(',A10,') = ',1PE9.2,
     1      ' Column Density = ',1PE9.2)
 2021 FORMAT(10X,' Kinetic temp. = ',0PF6.1)
 2030 FORMAT(10X,'(J,pop)',2X,4('(',I2,',',F6.4,')')) 
 2050 FORMAT (3X,6HJI  JF,7X,4HBETA,7X,10HEXCIT TEMP 
     1 ,6X,3HTAU,8X,8HRAD TEMP,7X,4HFREQ) 
 2060 FORMAT (3X,I2,2X,I2,2X,3X,F8.5,5X,F8.2,2X,2(3X,1PE10.3),3X, 
     1     0PF9.3) 
      RETURN 
      END
C ***********************************************************
C **************End of PRTOUT********************************
C ***********************************************************
      subroutine seltrans(trans,ntrans)
c
c   fill out logical array to select transitions to be written
c   to miriad data file based either user input transition list 
c   or default to auto selection of MAXTRAN transitions. In auto
c   mode all transition are selected if there are MAXTRAN or less;
c   if there are more, only transitions between 70 and 500 GHz are
c   selected -- up to 20 of them.
c
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      REAL trans(MAXTRAN),selfr(MAXTRAN)
      INTEGER ntrans,i,j,k,nselect
c
      if(ntrans .gt. 0) then
	 nselect = 0
	 do i=1,nlev
	 do j=1,nlev
	    if(a(i,j) .gt. 0.0) then
	       do k=1,ntrans
		  if(abs(trans(k)-freq(i,j)) .lt. 0.002) then
		     nselect = nselect+1
		     selfr(nselect) = freq(i,j)
		     wrtrans(i,j) = .true.
		  endif
	       enddo
	    else
               wrtrans(i,j) = .false.
	    endif
	 enddo
	 enddo
	 do i=1,nselect
	    trans(i) = selfr(i)
	 enddo
	 ntrans = nselect
	 if(ntrans .eq. 0) 
     1      call bug('f','No Transitions selected; check trans values')
      else
	 nselect = 0
	 do i=1,nlev
	 do j=1,nlev
	    if(a(i,j) .gt. 0.0) nselect = nselect + 1
	 enddo
	 enddo
	 if(nselect .le. 20) then
	    nselect = 0
	    do i=1,nlev
	    do j=1,nlev
	       if(a(i,j) .gt. 0.0) then
		  nselect = nselect + 1
		  trans(nselect) = freq(i,j)
		  wrtrans(i,j) = .true.
	       else
		  wrtrans(i,j) = .false.
	       endif
	    enddo
	    enddo
	 else
	    nselect = 0
            do i=1,nlev
            do j=1,nlev
               if(a(i,j) .gt. 0.0 .and. nselect .lt. 20) then
   	          if(freq(i,j) .gt. 70.0 .and. 
     1			      freq(i,j) .lt. 500.) then
	 	     nselect = nselect + 1
		     trans(nselect) = freq(i,j)
		     wrtrans(i,j) = .true.
	 	  else
		     wrtrans(i,j) = .false.
	 	  endif
	       endif
	    enddo
	    enddo
	 endif
      endif  
      return
      end
c*******************************************************************
c****************** end of seltran *********************************
c*******************************************************************
