	program lvgchi
c
c   Calculates the chi-square values for input observational line 
c   temperatures versus a grid of LVG models.
c
c  History:
c    lgm 03jul92   Beginning writing of program from scratch.
c    lgm 13jul92   updated documentation for keyword data
c    lgm 19sep92   corrected error in handling 3-D model cubes
c    lgm 01jun94   added logclose statement to end of main
c    lgm 06aug97   updated log writes to be compatable with next miriad subs
c    pjt 02jul02   ansi fixes for linux fortran
c
c------ Inline doc (retrieved with doc to a .doc file) ----------------c
c
c= LVGCHI - Calculates chi-square between data and model grid
c& lgm
c: models
c+
c  Calculates the chi-square values for input observational line
c  temperatures against a grid of LVG models. The models are
c  assumed to be generated by the program LVG.
c@ data
c   Name of file containing observational data in the following
c   format: # of freqs
c           freq1 freq2 freq3 ..... freqn
c           temp1 err1 temp2 err2 temp3 err3 ..... tempn errn
c           temp1 err1 temp2 err2 temp3 err3 ..... tempn errn
c           ..... .... ..... .... ..... .... ..... ..... ....
c   where all data on a single line are assumed to apply to a
c   single position. Each input line in the data file is assumed to 
c   be an independent position. All points must have measurements
c   at all input frequencies. Maximum of 100 positions and 20 
c   frequencies are allowed.
c   No default.
c@ model
c   Name of file containing model grid for chi-square calculation.
c   This model grid should have been generated by LVG and may be
c   either a 3-D or 4-D cube.
c   No default.
c@ out
c   Name of file for output chi-square cube. ONLY ONE data point
c   may be done if out is specified.
c@ log
c   Log file name. Default is to write to standard output.
c
c----------------------------------------------------------------
c
	INCLUDE 'rad_size.h'
	INCLUDE 'cubes.h'
c
	integer MAXDATA,MAXNAX 
	parameter (MAXDATA=200)
	parameter (MAXNAX=5)
	character datafile*50,modfile*50,logfile*50
	character outfile*50,version*25
	character line*100,words(30)*40,molcul*10
	real crpix(4)
	real datfreq(MAXTRAN),data(MAXDATA,MAXTRAN)
        real daterr(MAXDATA,MAXTRAN)
	real tranfr(MAXTRAN),chisq(MAXMOD,MAXMOD,MAXMOD)
	real axval(4),minchi,row(MAXMOD),newval
	real range(4,2)
	integer i,j,k,l,in,naxis,nwords,iplane,ndatfr
	integer idat,imin(3),maxplane,planeid(MAXTRAN)
	integer tout
	logical eof,more,outing
c
	data eof /.false. /
	data more /.true./
        DATA version / 'Version August 26, 2002' /
c
	call output('LVGCHI:'//version)
c
c  get user inputs
c
	call keyini
	call keya('data',datafile,'  ')
	if(datafile(1:2) .eq. '  ') 
     1			call bug('f','DATA must be input')
	call keya('model',modfile,'  ')
	if(modfile(1:2) .eq. '  ')
     1			call bug('f','MODEL must be input')
	call keya('out',outfile,' ')
	call keya('log',logfile,' ')
	call LogOpen(logfile,' ')
	call keyfin
c
c  open model file and get data
c
	nsize(4) = 1
	naxis    = 4
	call xyopen(in,modfile,'old',naxis,nsize)
	call rdhdi(in,'naxis',naxis,0)
	if(naxis .le. 2) call bug('f','NAXIS<3 in model image')
	call rdhead(in,naxis,molcul,ctyp,crval,crdel,crpix)
	call rdgrid(in,naxis)
c
c  read freq information from history file
c
	maxplane = 0
	call hisopen(in,'read')
	do while (.not. eof) 
	   call hisread(in,line,eof)
	   if(line(1:3) .eq. 'LVG') then
	      call sopchr(line,words,nwords)
	      if(words(2)(1:6) .eq. 'PLANE ') then
		 read(words(4),*) iplane
		 read(words(6),*) freq
		 tranfr(iplane) = freq
		 if(iplane .gt. maxplane) maxplane = iplane
	      endif
	   endif
	enddo
	call hisclose(in)
c
c  read in data file now
c
	open(unit=22,file=datafile,status='old')
	read(22,*) ndatfr
	read(22,*) (datfreq(i),i=1,ndatfr)
	do i=1,MAXDATA
	   read(22,*,end=105) (data(i,j),daterr(i,j),j=1,ndatfr)
	   ndata = i
	enddo
  105	continue
c
c  identify observational data frequencies with planes in cube
c
	do i=1,maxplane
           freq = tranfr(i)
	   do j=1,ndatfr
	      if(abs(datfreq(j)-freq) .lt. 0.001) planeid(j) = i
	   enddo
	enddo
	do i=1,ndatfr
	   if(planeid(i) .eq. 0) then
	      call bug('w',' Data freq not found in model cube')
	      write(line,'('' Data freq: '',f8.2)') datfreq(i)
	      call bug('f',line)
	   endif
	enddo
c
c  open the chisquare image file of specified and only one
c  point being done.
c
	outing = .false.
	if(outfile(1:1) .ne. ' ') then
	   if(ndata .eq. 1) then
	      outing = .true.
	      call xyopen(tout,outfile,'new',naxis-1,nsize)
              call wrhead(tout,naxis-1,nsize,ctyp,crpix,crval,
     1                  crdel,molcul)
	      call hisopen(tout,'write')
	      call hisappn(tout,modfile,.false.)
	      call hiswrite(tout,'LVGCHI:'//version)
	      call hisinput(tout,'chisqr')
	   else
	     call bug('f','One one data point allowed')
	   endif
	endif
c
c  form chi-square cube
c  
	if(naxis .eq. 4) then
	   nax3 = nsize(3)
	else
	   nax3 = 1
	endif
	do idat=1,ndata
	   do k=1,nax3
           do j=1,nsize(2)
           do i=1,nsize(1)
	      chisq(i,j,k) = 0.0
	   enddo
	   enddo
	   enddo
	   do ipl=1,ndatfr
	      l = planeid(ipl)
	      do k=1,nax3
	      do j=1,nsize(2)
	      do i=1,nsize(1)
		 newval = ((data(idat,ipl)-cube(i,j,k,l))**2)/
     1			  (daterr(idat,ipl)*daterr(idat,ipl))
		 chisq(i,j,k) = chisq(i,j,k) + newval
	      enddo
	      enddo
	      enddo
	   enddo
c
c   find minimum of cube
c
	   minchi = 1.0e6
	   do i=1,3
	     range(i,1) =  1.0e6
	     range(i,2) = -1.0e6
	   enddo
           do k=1,nax3
           do j=1,nsize(2)
           do i=1,nsize(1)
	      if(chisq(i,j,k) .lt. minchi) then
	         imin(1) = i
	         imin(2) = j
	         imin(3) = k
	         minchi = chisq(i,j,k)/(ndatfr-1)
 	      endif
	      if((chisq(i,j,k)/(ndatfr-1)) .lt. 1.0) then
	        if(i .lt. range(1,1)) range(1,1) = i
	        if(i .gt. range(1,2)) range(1,2) = i
                if(j .lt. range(2,1)) range(2,1) = j
	        if(j .gt. range(2,2)) range(2,2) = j
                if(k .lt. range(3,1)) range(3,1) = k
                if(k .gt. range(3,2)) range(3,2) = k
              endif
  	   enddo
	   enddo
	   enddo
c
c   write out results
c
           write(line,'(''Observational data point: '',i3)') idat
           call Logwrite(line,more)
	   if(outing) then
	      call hiswrite(tout,'LVGCHI: '//line)
	      do i=1,ndatfr
		 write(line,
     1	      '(''Freq: '',f8.2,'' Tr: '',f8.2,'' sigma: '',f7.3)')
     2			datfreq(i),data(1,i),daterr(1,i)
	         call hiswrite(tout,'LVGCHI: '//line)
	      enddo
	   endif
	   write(line,'(''   Minimum in Reduced Chi-sqr: '',f8.2, 
     1              ''  at model pixel: '',i3,1x,i3,1x,i3)') 
     2              minchi,(imin(i),i=1,3)
	   call Logwrite(line,more)
           if(outing) call hiswrite(tout,'LVGCHI: '//line)
	   do i=1,3
	    axval(i) = crval(i) + (imin(i)-crpix(i))*crdel(i)
	   enddo  
	   write(line,'(''   Best Fit: '',3(a,''= '',1pe9.3,3x))')
     1			(ctyp(i),(10.0**axval(i)),i=1,3)
           call Logwrite(line,more)
           if(outing) call hiswrite(tout,'LVGCHI: '//line)
           if(range(1,1) .lt. (nsize(1) + 1)) then
	     do ii=1,3
               axval(1) = crval(ii) + (range(ii,1)-crpix(ii))*crdel(ii)
               axval(2) = crval(ii) + (range(ii,2)-crpix(ii))*crdel(ii)
	       write(line,'(''       chi^2<1 range in '',a,'':'',
     1               2(1pe9.3,2x))') ctyp(ii),((10.0**axval(i)),i=1,2)
               call LogWrite(line,more)
               if(outing) call hiswrite(tout,'LVGCHI: '//line)
	     enddo
           else
	     write(line,'(''    chi^2 < 1 never achieved '')')
	     call Log Write(line,more)
             if(outing) call hiswrite(tout,'LVGCHI: '//line)
           endif
	enddo
c
c   write out the chi-square image if requested
c
	if(outing) then
           if(naxis-1 .eq. 3) then
              do k=1,nsize(3)
                 call xysetpl(tout,1,k)
                 do j=1,nsize(2)
                    do i=1,nsize(1)
                       row(i) = chisq(i,j,k)
                    enddo
                    call xywrite(tout,j,row)
                 enddo
              enddo
           else
	      k = 1
              call xysetpl(tout,1,1)
              do j=1,nsize(2)
                 do i=1,nsize(1)
                    row(i) = chisq(i,j,k)
                 enddo
                 call xywrite(tout,j,row)
              enddo
           endif
	   call hisclose(tout)
	   call xyclose(tout)
	endif

	call xyclose(in)
        call logclose
	stop
	end

	subroutine rdhead(in,naxis,molcul,ctyp,crval,crdel,crpix)
c
	character*(*) ctyp(4),molcul
	real crval(4),crdel(4),crpix(4)
	double precision dummy
c
	call rdhda(in,'object',molcul,' ')
	call rdhda(in,'ctype1',ctyp(1),' ')
        call rdhda(in,'ctype2',ctyp(2),' ')
        call rdhda(in,'ctype3',ctyp(3),' ')
	call rdhdd(in,'crpix1',dummy,0.0d0)
	crpix(1) = dummy
        call rdhdd(in,'crpix2',dummy,0.0d0)
        crpix(2) = dummy
        call rdhdd(in,'crpix3',dummy,0.0d0)
        crpix(3) = dummy
	call rdhdd(in,'cdelt1',dummy,0.0d0)
	crdel(1) = dummy
        call rdhdd(in,'cdelt2',dummy,0.0d0)
        crdel(2) = dummy
        call rdhdd(in,'cdelt3',dummy,0.0d0)
        crdel(3) = dummy
	call rdhdd(in,'crval1',dummy,0.0d0)
	crval(1) = dummy
        call rdhdd(in,'crval2',dummy,0.0d0)
        crval(2) = dummy
        call rdhdd(in,'crval3',dummy,0.0d0)
        crval(3) = dummy
	if(naxis .eq. 4) then
           call rdhda(in,'ctype4',ctyp(4),' ')
           call rdhdd(in,'crpix4',dummy,0.0d0)
           crpix(4) = dummy
           call rdhdd(in,'cdelt4',dummy,0.0d0)
           crdel(4) = dummy
           call rdhdd(in,'crval4',dummy,0.0d0)
           crval(4) = dummy
	endif
c
	return
	end
	
	subroutine rdgrid(in,naxis)
c
	INCLUDE 'rad_size.h'
        INCLUDE 'cubes.h'
c
	integer in,naxis,iax(2)
	real row(MAXMOD)
c
	if(naxis .eq. 4) then
	   do l=1,nsize(4)
	   do k=1,nsize(3)
	      iax(1) = k
	      iax(2) = l
	      call xysetpl(in,naxis-2,iax)
	      do j=1,nsize(2)
	         call xyread(in,j,row)
	         do i=1,nsize(1)
	            cube(i,j,k,l) = row(i)
	         enddo
	      enddo
	   enddo
	   enddo
	elseif (naxis .eq. 3) then
	   do l=1,nsize(3)
              iax(1) = l
              iax(2) = 1
              call xysetpl(in,naxis-2,iax)
              do j=1,nsize(2)
                 call xyread(in,j,row)
                 do i=1,nsize(1)
                    cube(i,j,1,l) = row(i)
                 enddo
              enddo
           enddo
	else
	   call bug('f','RDGRID: NAXIS M must be 3 or 4')
	endif
	return
	end
        subroutine wrhead(lout,naxis,nsize,ctyp,crpix,crval,
     1			crdel,molcul)
C
C  writes header for image cube
C
        integer lout,naxis,nsize(4)
	real crpix(4),crdel(4),crval(4)
        character*(*) molcul,ctyp(4)
C
        call wrhdi(lout,'naxis',naxis)
        call wrhdi(lout,'naxis1',nsize(1))
        call wrhdi(lout,'naxis2',nsize(2))
        call wrhda(lout,'ctype1',ctyp(1))
        call wrhda(lout,'ctype2',ctyp(2))
        call wrhdd(lout,'crpix1',dble(crpix(1)))
        call wrhdd(lout,'crpix2',dble(crpix(2)))
        call wrhdd(lout,'cdelt1',dble(crdel(1)))
        call wrhdd(lout,'cdelt2',dble(crdel(2)))
        call wrhdd(lout,'crval1',dble(crval(1)))
        call wrhdd(lout,'crval2',dble(crval(2)))
        call wrhda(lout,'object',molcul)
        if(naxis .eq. 3) then
           call wrhdi(lout,'naxis3',nsize(3))
           call wrhda(lout,'ctype3',ctyp(3))
           call wrhdd(lout,'crpix3',dble(crpix(3)))
           call wrhdd(lout,'cdelt3',dble(crdel(3)))
           call wrhdd(lout,'crval3',dble(crval(3)))
        endif
        return
        end
C
