       PROGRAM MICSPH 
C 
C   Calculates spherical microturbulent models for linear, asymmetric,
C   and symmetric top molecules.
C   ALL UNITS CGS 
c
c  History:
c    lgm  pre-his  Program created as an offshoot of a pre-miriad version
c    lgm  30jun92  first working and tested version
c    lgm  13jul92  compiled with larger MAXT value.
c    lgm  15jul92  Changed terp to save log(pops) internally and
c                    moved spline fitting into main iter loop
c    lgm  16jul92  Modified to use distance units of au. This cleaned
c                    up most of the underflow problems. A few remain.
c    lgm  01feb94  Corrected format error in temp printout, added call to
c                    logclose
c    lgm  06aug97   updated log writes to be compatable with next miriad subs
c                   corrected calls to SPLNC to use variable epsln
c    pjt  03jul02   fixed typo for ANSI fortran
C-------------------------------------------------------------------- 
c
c= MICSPH - Calculate spherical microturbulent models
c& lgm
c: models
c+
c    Calculates spherical microturbulent models for linear, asymmetric,
c    and symmetric top molecules.
c@ mol
c   Name of molecule. Known molecules are CO, CS, etc. No Default.
c@ iso
c   Isotope of molecule if other than main isotopic species. Default
c   is main isotopic species.
c@ nlayer
c   Number of radial layers in spherical model, specification of input
c   mode (either '  ' or 'file'), and input data file name (if mode='file'). 
c   The number of layers can be from 2 to MAXMOD (currently 20). 
c   The input mode for model radii, densities, fractional abundances,
c   and kinetic temperatures can be either power-law specification
c   using the key words below or a file specifying each in the
c   format (Radius  Density  Xmol  Tkin):
c           2.9250e16 6.60E6 4.0e-10 98.4
c           3.7021e16 4.86e6 4.0e-10 45.5
c   If file input is selected, the keywords rad, den, xmol, and tkin are
c   ignored.
c   Default: nlayer = 3.
c@ rad
c   Two numbers representing the inner and outer radius of the model plus
c   either log or lin indicating how your want the models distributed.
c   If only one number is entered, that value is used for the inner
c   radius and the outer radius is nlayer time the inner radius. 
c   Default: rad = 3.0e16, nlayer*3.0e16, lin
c@ den
c   Two numbers representing the densities at the inner radius and
c   the power law index for the relation: n = n(rinner)*(r/rinner)**a.
c   If only one number is entered, the index is assumed to be zero.
c   The densities at model radii between the inner and outer radii are 
c   interpolated assuming the power law dependence.
c   Default: den = 1.0e4,0.0
c@ xmol
c   Two numbers representing the molecular fractional abundance at
c   the inner radii and the power law inden for the relation:
c   x = x(rinner)*(r/rinner)**a. If only one number is entered, 
c   the index is assumed to be zero. Fractional abundances
c   for model radii between the inner and outer radii are interpolated
c   assuming the power law dependence.
c   Default: xmol = 1.0e-9,0.0
c@ tkin
c   Two numbers representing the gas temperature at the inner radii and
c   the power law index for the relation: T = T(rinner)*(r/rinner)**a. 
c   If only one number is entered, the index is assumed to be zero.
c   The temperature is interpolated as a power law between the inner 
c   and outer radii. 
c   Default: tkin = 20.0,0.0
c@ hole
c   Logical indicating whether the density should be set to zero inside
c   the inner radius (true) or extrapolated inward with the same value 
c   as at the inner radius (false). 
c   Default is false.
c@ dvel
c   Doppler velocity line width (FWHM) in km/sec. 
c   Default: dvel = 1.0
c@ tbkg
c   Blackbody temperature for background radiation field in all transitions.
c   This is normally set to correspond to the cosmic background temperature.
c   Default: tbkg = 2.78.
c@ conv
c   Fractional change criterion applied to populations to decide if the 
c   model has converged to a solution. This number is compared to
c   the quantity [(NEW POP)-(OLD POP)]/(NEW POP) for each molecular
c   level and each layer. If change is too large, another iteration is
c   done. Allowed range of conv is from 0.1 to 0.0001.
c   Default: conv = 0.005
c@ iter
c   Maximum number of iterations allowed. If the model does not converge 
c   in this number of iterations, the calculation is stopped and the
c   results printed with a warning message about the non-convergence
c   Default: iter = 25
c@ pops
c   Name of miriad file were  level populations for each radii are
c   to be written.
c   No default.
c@ print
c   Parameter which control level of printed output.
c      0 = minimum output which is useful when running grids
c      1 = normal output level
c      2 = detailed output level - used for debugging and error checking
c@ log
c   Name of log file were full printed output for human consumption
c   is written. 
c   Default is to write to screen.
c
c-------------------------------------------------------------------
c  include files
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      INCLUDE 'radinit.h'
      INCLUDE 'mic_tau.h'
      INCLUDE 'mic_radsup.h'

      CHARACTER*10 MOLCUL,MOLD,ISOTOP
      CHARACTER LINE*80,popname*50,logfile*50,loglin*10
      CHARACTER mode*10,infile*50,version*25
      INTEGER I,INAM,IMOL,iprint,pout
      REAL CONVRJ,RSTEP
      REAL den1,xmol1,tk1,rad1,dlaw,xlaw,tlaw
      LOGICAL table,more
C
      DATA MOLD / 'OLDNAME   ' /
      DATA more /.true. /
      DATA version / ' Version July 3 2002  ' /
C
      CALL output('MICSPH:'//version)

C------------------------------------------------------------
C   Get input using key routines
C
      CALL KEYINI
      CALL KEYA('mol',molcul,' ')
      if(molcul .eq. ' ') call bug('f','Molecule must be specified')
      CALL KEYA('iso',isotop,molcul)
C
C    Input and check number of shells
C
      CALL KEYI('nlayer',npts,3)
      if(npts .lt. 2)   npts = 2
      if(npts .gt. MAXMOD) then
         call bug('w',' Number of models reset to max allowed')
         npts = MAXMOD
      endif
      CALL KEYA('nlayer',mode,' ')
      table = .false.
      if(mode(1:2) .eq. 'fi') then
         call keya('nlayer',infile,' ')
         if(infile(1:2) .eq. '  ') then
            call bug('f','Data input file must be specified')
         else
            open(27,file=infile,status='old')
            table = .true.
         endif
      endif
C
C   Input and check radii for calculation
C
      CALL KEYR('rad',rad1,3e16)
      CALL KEYR('rad',rmax,npts*rad1)
      CALL KEYA('rad',loglin,'lin')
      if(rad1 .le. 1.0e10) 
     1		call bug('f','Inner raduis must be > 1.0e10')
      if(rad1 .gt. rmax) 
     1		call bug('f','Inner radius must be < outer radius')
      if(rmax .gt. 3.0e21)
     1		call bug('f','Get real ! Outer radius too large')
      rad1 = rad1/au
      rmax = rmax/au
C
C   Input and check densities
C
      CALL KEYR('den',den1,1.0e4)
      CALL KEYR('den',dlaw,0.0)
      if(den1 .lt. 0.0) 
     1     call bug('f',' Density <= 0.0 not allowed')
C
C   Input and check fractional abundances
C
      CALL KEYR('xmol',xmol1,1.0e-9)
      CALL KEYR('xmol',xlaw,0.0)
      if(xmol1 .gt. 0.99) 
     1	   call bug('f','Fract. abundance must be less than 1')
C
C   Input and check temperatures
C
      CALL KEYR('tkin',tk1,20.0)
      CALL KEYR('tkin',tlaw,0.0)
      if(tk1 .lt. 0.9) 
     1	   call bug('f','Kinetic temp must be greater than 1K')
C
C   Input velocity width
C
      CALL KEYR('dvel',dopv,1.0)
c
c   Make a central hole in the sphere?
c
      CALL keyl('hole',hole,.false.)
C
C   input and check bacground temp
C
      CALL KEYR('tbkg',tbk,2.78)
      if(tbk .gt. 1.0e4 .or. tbk .lt. 0.1) 
     1     call bug('f','TBGK must be between 0.1 to 1e4')
C
C   input and check convergence criterion
C
      CALL KEYR('conv',convrj,0.005)
      if(convrj .gt. 0.5) then
         call bug('w','Convergence Criterion reset to 0.1')
         convrj = 0.1
      elseif(convrj .lt. 0.0001) then
         call bug('w','Convergence Criterion reset to 0.0001')
         convrj = 0.0001
      endif
C
C   Input and check iter 
C
      CALL KEYI('iter',nit,25)
      if(nit .gt. 1000 .or. nit .le. 0) then
         call bug('w','ITER out of range 1 to 1000; Reset to default')
         nit = 70
      endif
C
C   Get name for output miriad image containing population and model
C   info
C
      CALL KEYA('pops',popname,' ')
      if(popname .eq. ' ') 
     1	   call bug('f',' Output file must be specified')
C
C   Input print level specification
C
      CALL KEYI('print',iprint,1)
C
C   Get specification of log file 
C
      CALL KEYA('log',logfile,' ')
      
      CALL KEYFIN
C
C   End of input from key routines
C------------------------------------------------------------
C   Open log file of specified
C 
      call logopen(logfile,' ')
C-----------------------------------------------------------
      if(loglin(1:3) .eq. 'log') then
         write(line,'('' Using Log steps in radius'')')
         call LogWrite(line,more)
         rstep = alog(0.9995*rmax/rad1)/float(npts-1)
      else
         write(line,'('' Using Linear steps in radius'')')
         call LogWrite(line,more)
         rstep = 0.999*(RMAX-RAD1)/FLOAT(NPTS-1)
      endif
C---------------------------------------------------------------
C   Fill in density, fractional abundance, column density, and
C   temperature arrays according to specification
C
      if(table) then
         read(27,'(a)') line
         call LogWrite('Reading layer information from file',more)
         do i=1,npts
            read(27,*) rpt(i),rnh2(i),rnx(i),rtk(i)
	    rpt(i) = rpt(i)/au
            rnh2(i) = alog(rnh2(i))
            rnx(i)  = alog(rnx(i))
         enddo
	 rmax = 1.0001*rpt(npts)
      else
         do i=1,npts
            if(loglin(1:3) .eq. 'log') then
               rpt(i) = rad1 * exp(rstep*(i-1))
            else
               rpt(i) = rstep*FLOAT(I-1) + rad1
            endif
            RNH2(I) = ALOG(den1) + dlaw*alog(rpt(i)/rad1)
            RNX(I)  = ALOG(xmol1) + xlaw*alog(rpt(i)/rad1)
            RTK(I)  = tk1*((rpt(i)/rad1)**tlaw)
         enddo
      endif
C------------------------------------------------------
C  This section writes out the model parameters to the
C  chosen output location
C
      WRITE(LINE,2000)
      CALL LogWrite(line,more)
      WRITE(line,2010) ISOTOP
      CALL LogWrite(line,more)
      if(table) then
         call LogWrite('Model specified as an input table',more)
         call LogWrite(' Radius  Density   Xmol Tkin',more)
         do i=1,npts
            write(line,'(1pe10.3,2(2x,e10.3),2x,0pf6.2)') 
     1                au*rpt(i),exp(rnh2(i)),exp(rnx(i)),rtk(i)
	    call LogWrite(line,more)
         enddo
      else
         WRITE(line,2020) au*RAD1,au*RMAX
         call LogWrite(line,more)
         IF(abs(rnh2(1)-rnh2(npts)) .gt. 0.001) THEN
            WRITE(line,2030) EXP(RNH2(1)),NPTS,EXP(RNH2(NPTS))
            call LogWrite(line,more)
            WRITE(line,2070) dlaw
         ELSE
            WRITE(line,2035) EXP(RNH2(1))
         ENDIF
         call LogWrite(line,more)
         IF(abs(rnx(1)-rnx(npts)) .gt. 0.001) THEN
            WRITE(line,2050) ISOTOP,EXP(RNX(1)),NPTS,
     1			EXP(RNX(NPTS))
            call LogWrite(line,more)
            WRITE(line,2070) xlaw
         ELSE
            WRITE(line,2055) ISOTOP,EXP(RNX(1))
         ENDIF
         call LogWrite(line,more)
         IF(abs(rtk(1)-rtk(npts)) .gt. 0.01) THEN
            WRITE(line,2060)RTK(1),NPTS,RTK(NPTS)
            call LogWrite(line,more)
            WRITE(line,2070) tlaw
         ELSE
            WRITE(line,2065) RTK(1)
         ENDIF
         call LogWrite(line,more)
      endif
      WRITE(line,2080) DOPV
      call LogWrite(line,more)
      WRITE(line,2090) TBK
      call LogWrite(line,more)
      WRITE(line,2095) NIT,CONVRJ
      call LogWrite(line,more)
 2000 FORMAT(20X,'SPHERICALLY SYMMETRIC MICROTURBULENT MODELS')
 2010 FORMAT(2X,16HINPUT QUANTITIES,6X,'Molecule name = ',A10)
 2020 FORMAT(5X,'Inner radius = ',1PE9.2,'  Outer radius = ',E9.2)
 2030 FORMAT(5X,'Density at model 1 = ',1PE9.2,2X,'at model ',I2,
     1        ' = ',1PE9.2)
 2035 FORMAT(5X,'Density fixed at ',1PE9.2,' for all models')
 2050 FORMAT(5X,'X(',A10,') at model 1 = ',1PE9.2,2X,'at model ',I2,
     1        ' = ',1PE9.2)
 2055 FORMAT(5X,'X(',A10,') fixed at ',1PE9.2,' for all models')
 2060 FORMAT(5X,'T kin at model 1 = ',F5.1,2X,'at model ',I2,' = ',
     1        F5.1)
 2065 FORMAT(5X,'T kin fixed at ',F5.1,' for all models')
 2070 FORMAT(15X,'Power law relation: power = ',f6.2)
 2080 FORMAT(5X,'Doppler Line Width (FWHM) = ',F6.2,' Km/s')
 2090 FORMAT(5X,'Background Temperature fixed at ',F6.2)
 2095 FORMAT(5X,'Maximum # iterations = ',I4,' Fract. Converg. = ',
     1		F7.4)
C 
C End of loop for entering and translating model parameters
C------------------------------------------------------------ 
C  This area of code reads in the collision rates and, for asymmetric tops,
C  reads in the Einstein coefficients. For linear molecules
C  IMOL=1; for asymmetric molecules, IMOL=-1. If isotopic
C  corrections to freqs or E values are necessary, they are done here.
C
      call LogWrite(' ',more)
      CALL MOLREAD(MOLCUL,INAM,IMOL)
      IF(INAM.EQ.0) CALL bug('f','Molecule not found')
      IF(MOLD.NE.MOLCUL) THEN
         CALL ISOID(ISOTOP,IMOL)
         DYOLD = DY
         CALL COLL1(IPRINT,ISOTOP,IMOL)
         MOLD=MOLCUL
      ENDIF
      call LogWrite(' ',more)
      IF(IMOL.EQ.1) THEN
         WRITE(line,2100) BROT,DE
      ELSEIF(IMOL .eq. -1) THEN
         WRITE(line,2110) A0,B0,C0
      ELSE
         WRITE(line,2120) B0,C0,DJ
      ENDIF
      call LogWrite(line,more)
      WRITE(line,2130) DY
      call LogWrite(line,more)
 2100 FORMAT(5X,'Rotational Constant = ',1PE10.3,3X,
     1       'Distortion Constant = ',E10.3)
 2110 FORMAT(5X,'A0 = ',1PE10.3,2X,'B0 = ',E10.3,2X,'C0 = ',E10.3)
 2120 FORMAT(5X,'B0 = ',1Pe10.3,2x,'C0 = ',E10.3,2x,'DJ = ',E10.3)
 2130 FORMAT(5X,'Dipole moment = ',F4.2,' Dbye')
C-------------------------------------------------------------------
C  The molecule radiative coefficients, energy levels, and level
C  degeneracies are calculated in RAD---. COLL2 interpolates the
C  collision rates for the temperature of the first model.
C
      CALL COLL2(RTK(1))
      DYOLD=DY
C-----------------------------------------------------------------------
C  Go do all of the work in iter
C-----------------------------------------------------------------------
      CALL ITER(CONVRJ,MOLCUL) 
      CALL writeout(popname,molcul,pout)
      CALL TROUT (MOLCUL)
C
C  write out additional model info to history file
C
      call hisopen(pout,'append')
      call hiswrite(pout,'MICSPHR:'//version)
      call hisinput(pout,'micsph')
      write(line,'(''MICSPH: MODEL nmodels = '',i4)') npts
      call hiswrite(pout,line)
      write(line,'(''MICSPH: MODEL iso = '',a)') isotop
      call hiswrite(pout,line)
      write(line,'(''MICSPH: MODEL SHELL  DENSITY  XMOL  TKIN'')')
      call hiswrite(pout,line)
      do i=1,npts
	 write(line,'(''MICSPH: MODEL '',4(e12.4,2x))') au*rpt(i),
     1			exp(rnh2(i)),exp(rnx(i)),rtk(i)
         call hiswrite(pout,line)
      enddo
      if(hole) then
         write(line,'(''MICSPH: MODEL Hole: true'')') 
      else
         write(line,'(''MICSPH: MODEL Hole: false'')')
      endif
      call hiswrite(pout,line)
      write(line,'(''MICSPH: MODEL dvel = '',f7.2)') dopv
      call hiswrite(pout,line)
      write(line,'(''MICSPH: MODEL tbkg = '',f8.2)')  tbk
      call hiswrite(pout,line)
      call hisclose(pout)
      call xyclose(pout)
      call logclose

      IF(OUTC.EQ.9) CLOSE(UNIT=OUTC)
      CLOSE(UNIT=12)
      STOP 
      END 
C*******************************************************************
C********************* END OF MAIN *********************************
C*******************************************************************
      SUBROUTINE ITER(CONVRJ,MOLCUL) 
C
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      INCLUDE 'mic_radtr.h'
      INCLUDE 'mic_tau.h'
C
      DOUBLE PRECISION XPOPS(MAXLEV),STEQ(MAXLEV,MAXLEV),SUM
      INTEGER ITT,I,II,IPT,IL,NWAR,N,J,I1,J1
      INTEGER nn,kl,ku,kk,iflag
      REAL TOLD,DELTAU,den,epsln
      REAL DUM,XDIF,DIF,DIFR,CONVRJ,RPTH2,RPTX,X1,TEX
      CHARACTER MOLCUL*10,LINE*80
      LOGICAL more
C
      DATA more /.true. /
C
      ITT = 0 
      TOLD=0.0 
      DO 100 II=1,NPTS
         CALL XLTE(XPOPS,RTK(II))
         DO 90 I=1,NLEV 
            POPN(I,II)=XPOPS(I) 
   90    CONTINUE
  100 CONTINUE
      epsln = 1.0e-6
      call splnc(npts,rpt,rnh2,epsln,snh2)
      call splnc(npts,rpt,rnx,epsln,sxm)
C ****************************************************** 
C              MAIN ITTERATION LOOP STARTS HERE 
C      XINT -
C      STAT - set-up statistical eq. equations
C      SOLV - solve stat.eq. by matrix inversion
C ****************************************************** 
   55 CONTINUE 
      NWAR=0
      XDIF=0.0
      iflag = 0
      DO 300 IPT=1,NPTS 
         CALL TERP(POPN,iflag,DUM)
	 CALL COLL2(rtk(ipt))
         DEN = EXP(RNH2(IPT))
	 DO 210 I=1,NLEV
            XN(I) = POPN(I,IPT)
  210    CONTINUE
         CALL XINT(IPT,ITT) 
         CALL STATEQ(DEN,XJ,STEQ,MAXLEV) 
         CALL SOLV(STEQ,XN,NLEV,XPOPS) 
C
C    Normalize level population for zone IPT
C
         SUM=0.0 
         DO 220 I=1,NLEV 
            SUM=SUM+XPOPS(I) 
  220    CONTINUE
         DO 230 I=1,NLEV 
            POPN(I,IPT)=XPOPS(I)/SUM 
  230    CONTINUE
C 
C    For first iteration on each zone propagate out level
C    populations from next level in
C
         IF(ITT.EQ.0) THEN 
	    IF(IPT.LT.NPTS) THEN
               DO 240 IL=1,NLEV 
	          POPN(IL,IPT+1)=POPN(IL,IPT) 
  240          CONTINUE 
	    ENDIF
            CALL TERP(POPN,iflag,DUM) 
         ENDIF
C
         DO 270 I=1,NLEV-1
            DIF = ABS(POPN(I,IPT)-XN(I))
            IF(XN(I).GT.1.0E-8) THEN
               DIFR=DIF/XN(I)
               IF(DIFR .GT. CONVRJ) NWAR=1
               IF(DIFR .GT. XDIF) XDIF=DIFR
            ENDIF
  270    CONTINUE
  300 CONTINUE 
      ITT=ITT+1 
      IF(ITT.EQ.1) GO TO 55 
      IF(NWAR .EQ. 0) GO TO 385 
      IF(ITT.LT.NIT) GO TO 55 
      CALL LogWrite(' Did not converge',more) 
      WRITE(line,2110) XDIF 
      call LogWrite(line,more)
 2110 FORMAT(1X,40HLARGEST FRACTIONAL CHANGE LAST ITERATION,1X,F9.4) 
  385 CONTINUE 
C ********************************************* 
C MAIN ITTERATION LOOP ENDS HERE 
C ********************************************* 
      CALL TERP(POPN,iflag,DUM)
      call LogWrite(' ',more)
      call LogWrite(' ',more)
      WRITE(line,2000) NPTS,ITT 
      call LogWrite(line,more)
      call LogWrite(' ',more)
      call LogWrite(' ',more)
      DO 500 N=1,NPTS 
	 RADIUS= RPT(N)
         RPTH2 = EXP(RNH2(N)) 
         RPTX  = EXP(RNX(N)) 
         call LogWrite(' ',more)
         call LogWrite(' ',more)
         WRITE(line,2020) au*RPT(N) 
	 call LogWrite(line,more)
         WRITE(line,2025) RPTH2,MOLCUL,RPTX,RTK(N)
         call LogWrite(line,more)
         NN=(NLEV/4)+1
         DO 490 I=1,NN
            KL=(I-1)*4+1
            IF(KL.GT.NLEV) GO TO 490
            KU=KL+3
            IF(KU.GT.NLEV) KU=NLEV
            WRITE(line,2030) (kk-1,POPN(kk,N),kk=kl,ku)
            call LogWrite(line,more)
  490    CONTINUE
         call LogWrite(' ',more)
         WRITE(line,2040) 
         call LogWrite(line,more)
 2000 FORMAT(4X,I4,' LAYERS CONVERGED IN ',I2,1X,11HITTERATIONS) 
 2020 FORMAT(27X,6HRADIUS,1X,1PE10.3)
 2025 FORMAT(12X,7HN(H2)= ,1PE9.3,3X,2HX(,A5,3H)= ,E9.3,3X,
     1    4HTK= ,0PF5.1) 
 2030 FORMAT(1X,'(J,POPS)',2X,4('(',I2,',',1PE11.4,')',1X))
 2040 FORMAT(8X,2HJU,2X,2HJL,9X,4HBETA,7X,7HDTAU/DR,9X,3HTEX,8X, 
     1    9HFREQ(GHZ)) 
         DO 450 I=2,NLEV 
         DO 450 J=1,NLEV 
            IF(A(I,J) .GT. 0.0) then 
               X1=(POPN(I,N)*DEGEN(J)/(POPN(J,N)*DEGEN(I))) 
               IF(X1 .GT. 0.) then 
                  TEX=-AH*FREQ(I,J)/(AK*ALOG(X1)) 
	       else
    	          TEX=0.0 
	       endif
               CALL CALDT(I,J,N,DELTAU)
               I1=I-1 
               J1=J-1 
               WRITE(line,2050) I1,J1,BETA(I,J,N),DELTAU,TEX,FREQ(I,J) 
               call LogWrite(line,more)
	    endif
  450    CONTINUE 
  500 CONTINUE 
 2050 FORMAT(8X,I2,2X,I2,5X,0PF9.5,4X,1PE10.3,4X,0PF9.3,5X,0PF9.3) 
      RETURN 
      END 
C******************************************************************
C******************** END OF ITER *********************************
C******************************************************************
      SUBROUTINE XINT(IPTT,ITT) 
C
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      INCLUDE 'mic_radtr.h'
      INCLUDE 'mic_radsup.h'
      INCLUDE 'mic_tau.h'
C
      REAL soutrm,esp,dtau,sterm,epray,gauss
      INTEGER IPTT,ITT,IM,I,iflag
      REAL CSOU,S,Y1,TMX,SUM,T0,T1,TSEC
      REAL RSEC,SOU1,TSMAL,BET1,S1,S2,RIPT,DTU
      REAL XT(5),YT(5),TSTP,epsln
      DOUBLE PRECISION STIMCOR
C
      EXTERNAL SOUTRM,ESP,DTAU,STERM,EPRAY,GAUSS
C
      DATA GAUMU / -.148874,0.148874,-.433395,0.433395,-.679410, 
     1   0.679410,-.865063,0.865063,-.973907,0.973907 /

      IPT  = IPTT 
      CSOU = 2.0E+18*ah/(AC*AC) 
      RIPT  = RPT(IPT)
      RADIUS = RIPT
      DO 500 JUP=2,NLEV 
      DO 500 JLO=1,NLEV 
         IF(A(JUP,JLO).EQ.0.0) GO TO 500 
         IF(ITT.LE.3) GO TO 55 
         IF(BETA(JUP,JLO,IPT).LT.0.998) GO TO 55 
         STIMCOR=POPN(JLO,IPT)*DEGEN(JUP)/(POPN(JUP,IPT)*
     1                DEGEN(JLO)) - 1.0D0
         IF(ABS(STIMCOR).LT.1.0E-8) STIMCOR = SIGN(1.0E-8,STIMCOR) 
         S=CSOU*(FREQ(JUP,JLO)**3)/STIMCOR 
         Y1=(1.0-BETA(JUP,JLO,IPT))*S 
         GO TO 495 
   55    CONTINUE 
         DO 200 IM=1,10 
            RMU=GAUMU(IM) 
            TMX=SQRT(RMAX*RMAX+(RMU*RMU-1.0)*RIPT*RIPT
     1                - RIPT*RMU) 
	    tstp   = tmx/(MAXT-1)
            SUM=0.0 
            AT(1)=0.0 
            ATAU(1)=0.0 
            DO 160 I=2,MAXT 
               T0     = tstp*(I-2) 
               T1     = tstp*(I-1)
               DTU    = GAUSS(T0,T1,3,DTAU) 
               SUM    = SUM+DTU
               AT(I)  = T1
               ATAU(I)= SUM
  160       CONTINUE
            epsln = 1.0e-6
            CALL SPLNC(MAXT,AT,ATAU,epsln,AA) 
            BETAMU(IM)=EPRAY(SUM) 
            TSEC=0.02*TMX
            RSEC=SQRT(RIPT*RIPT+TSEC*TSEC+2.0*RIPT*TSEC*RMU) 
            iflag = 1
            CALL TERP(POPN,iflag,RPT(IPT)) 
            SOU1=SOUR 
            CALL TERP(POPN,iflag,RSEC) 
            TSMAL=0.01*TSEC/(ABS((SOU1-SOUR)/SOU1) + 2.0e-3) 
	    IF(TSMAL .LT. 1.0e-5*TSEC) TSMAL = 1.0e-5*TSEC
            XT(1)=TSMAL 
            CALL SPLN(MAXT,1,AT,ATAU,XT,YT,AA) 
            BET1=EPRAY(YT(1)) 
            S1=(1.0-BET1)*SOU1 
            S2=GAUSS(TSMAL,TMX,6,STERM) 
            STAU(IM)=S1+S2 
            IF(RPT(IPT).LT.1.0E-6*RMAX) GO TO 485 
  200    CONTINUE 
         Y1=0.5*GAUSS(-1.0,1.0,10,SOUTRM) 
         BETA(JUP,JLO,IPT)=0.5*GAUSS(-1.0,1.0,10,ESP) 
         GO TO 495 
  485    Y1=STAU(1) 
         BETA(JUP,JLO,IPT)=BETAMU(1) 
  495    CONTINUE 
         XJ(JUP,JLO)=BETA(JUP,JLO,IPT)*BBK(JUP,JLO)+Y1 
  500 CONTINUE 
      RETURN 
      END 
C********************************************************************
C************************* END OF XINT ******************************
C********************************************************************
      FUNCTION SOUTRM(XMU) 
      INTEGER I,IM
      REAL XMU,SOUTRM
C
      INCLUDE 'mic_radsup.h'
C
      DO 100 I=1,10 
         IM=I 
         IF(ABS(XMU-GAUMU(I)).LT.1.0E-3) GO TO 105 
  100 CONTINUE 
      call bug('f',' FUNCTION STAU - BAD MU') 
  105 SOUTRM=STAU(IM) 
      RETURN 
      END 
C*********************************************************************
C************************** END OF SOUTRM ****************************
C*********************************************************************
      FUNCTION DTAU(TD) 
C
C   CALCULATES THE INCREMENTAL TAU AT THE DISTANCE TD ALONG AT RAY
C   AT ANGLE RMU ORIGINATING FROM RADIUS 
C
C   RVAL IS THE TRUE RADIUS CORRESPONDING TO THE POINT OF DISTANCE
C   TD ALONG THE RAY UNDER CONSIDERATION
C
      REAL CTAU,RIPT,TD,RVAL,DH2,XMOL,DTAU
      REAL XT(5),YT(5)
      INTEGER iflag
C
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      INCLUDE 'mic_radtr.h'
      INCLUDE 'mic_radsup.h'
      INCLUDE 'mic_tau.h'
C
C
c      CTAU=1.6651E-9*AH*AC*AU/(4.0*PI*RPI*1.0E5*DOPV) 
      CTAU = 2.22181e-9/dopv
      RIPT = RADIUS 
      RVAL = SQRT(RIPT*RIPT+TD*TD+2.0*RIPT*TD*RMU) 
      XT(1)=RVAL 
      IF(RVAL .LE. RPT(NPTS)) THEN
         CALL SPLN(NPTS,1,RPT,RNH2,XT,YT,SNH2) 
         DH2  = EXP(YT(1)) 
      ELSE
         DH2=0.0 
      ENDIF
      CALL SPLN(NPTS,1,RPT,RNX,XT,YT,SXM) 
      XMOL = EXP(YT(1)) 
      if(rval .lt. rpt(1) .and. hole) then
         dh2  = 0.0
         xmol = 0.0
      endif
      iflag = 1
      CALL TERP(POPN,iflag,RVAL) 
C      DTAU = CTAU * DH2 * XMOL * (XLO*BA(JLO,JUP) - XUP*BE(JUP,JLO))
      DTAU = DH2 * XMOL * (XLO*BA(JLO,JUP) - XUP*BE(JUP,JLO))
      if(abs(dtau) .gt. 1.0e-22) then
	 dtau = ctau*dtau
      else
	 dtau = 0.0
      endif
      RETURN 
      END 
C*******************************************************************
C*************************** END OF DTAU ***************************
C*******************************************************************
      FUNCTION SOUDTAU(TD)
C
C   CALCULATES THE INCREMENTAL SOURCE FUNCTION * dTAU AT THE DISTANCE TD ALONG AT RAY
C   AT ANGLE RMU ORIGINATING FROM RADIUS
C
C   RVAL IS THE TRUE RADIUS CORRESPONDING TO THE POINT OF DISTANCE
C   TD ALONG THE RAY UNDER CONSIDERATION
C
      REAL CTAU,RIPT,TD,RVAL,DH2,XMOL,DTAU
      REAL XT(5),YT(5),SOU,SOUDTAU
      INTEGER iflag
C
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      INCLUDE 'mic_radtr.h'
      INCLUDE 'mic_radsup.h'
      INCLUDE 'mic_tau.h'
C
C
C      CTAU=1.6651E-9*AH*AC*AU/(4.0*PI*RPI*1.0E5*DOPV)
      CTAU = 2.22181e-9/dopv
      RIPT = RADIUS
      RVAL = SQRT(RIPT*RIPT+TD*TD+2.0*RIPT*TD*RMU)
      XT(1)=RVAL
      IF(RVAL .LE. RPT(NPTS)) THEN
         CALL SPLN(NPTS,1,RPT,RNH2,XT,YT,SNH2)
         DH2  = EXP(YT(1))
      ELSE
         DH2=0.0
      ENDIF
      CALL SPLN(NPTS,1,RPT,RNX,XT,YT,SXM)
      XMOL = EXP(YT(1))
      if(rval .lt. rpt(1) .and. hole) then
         dh2  = 0.0
         xmol = 0.0
      endif
      iflag = 1
      CALL TERP(POPN,iflag,RVAL)
c      DTAU = CTAU * DH2 * XMOL * BA(JLO,JUP)/DEGEN(JUP)
      DTAU = DH2 * XMOL * BA(JLO,JUP)/DEGEN(JUP)
      SOU  = 2.0E+18*AH*(FREQ(JUP,JLO)**3)/(AC*AC) * 
     1			(XUP * DEGEN(JLO))
      SOUDTAU = SOU*DTAU
      if(abs(soudtau) .gt. 1.0e-22) then
	 soudtau = ctau*soudtau
      else
	 soudtau = 0.0
      endif
      RETURN
      END
C*******************************************************************
C************************** END OF SOUDTAU *************************
C*******************************************************************
      SUBROUTINE CALDT(ILEV,JLEV,N,DELTAU)
      INTEGER ILEV,JLEV,N
      REAL DELTAU,dtau
      EXTERNAL dtau
C
      INCLUDE 'mic_radsup.h'
C
      IPT    = N
      JUP    = ILEV
      JLO    = JLEV
      RMU    = 1.0
      DELTAU = DTAU(0.0)
      RETURN
      END
C*******************************************************************
C************************** END OF CALDT ***************************
C*******************************************************************
      FUNCTION STERM(TD) 
      REAL SDT,TD,STERM,SOUDTAU,sray
      REAL TNEW(5),TAUN(5)
      EXTERNAL SOUDTAU,SRAY
C
      INCLUDE 'rad_size.h'
      INCLUDE 'mic_radtr.h'
      INCLUDE 'mic_radsup.h'
      INCLUDE 'mic_tau.h'
C
      SDT     = SOUDTAU(TD) 
      TNEW(1) = TD 
      CALL SPLN(MAXT,1,AT,ATAU,TNEW,TAUN,AA) 
      STERM   = SDT*SRAY(TAUN(1)) 
      RETURN 
      END 
C********************************************************************
C*************************** END OF STERM ***************************
C********************************************************************
      FUNCTION SRAY(TAU) 
C INPUT PARAMETER IS TAU AT LINE CENTER 
C FUNCTION CALCULATES FREQ INTEGRAL FOR SOURCE TERM IN J 
C
      INTEGER K
      REAL TAU,SUM,TERM,SRAY
C
      IF(TAU.GT.7.0) GO TO 205 
      SUM=1.0/1.4142136-TAU/1.732051 
      TERM=-TAU/1.732051 
      DO 100 K=2,30 
         TERM=-TERM*TAU/FLOAT(K)*(FLOAT(K+1)/FLOAT(K+2))**0.5 
         SUM=SUM+TERM 
         IF(ABS(TERM).LT.1.0E-7) GO TO 105 
  100 CONTINUE 
  105 SRAY=SUM 
      RETURN 
  205 CONTINUE 
      IF(TAU.GT.15.0) GO TO 215 
      SRAY=(1.0819+1.618E-3*(TAU-11.0)*(TAU-11.0))/TAU**2.384 
      RETURN 
  215 CONTINUE 
      IF(TAU.GT.30.0) GO TO 225 
      SRAY=(0.7098+8.363E-5*(TAU-22.5)*(TAU-22.5))/TAU**2.222 
      RETURN 
  225 CONTINUE 
      SRAY=(0.59745+1.156E-5*(TAU-40.0)*(TAU-40.0))/TAU**2.17 
      RETURN 
      END 
C*********************************************************************
C************************* END OF SRAY *******************************
C*********************************************************************
      FUNCTION ESP(XMU) 
      INTEGER I,IM
      REAL XMU,ESP
C
      INCLUDE 'mic_radsup.h'
C
      DO 50 I=1,10 
         IM=I 
         IF(ABS(XMU-GAUMU(I)).LT.1.0E-3) GO TO 105 
   50 CONTINUE
      call bug('f',' FUNCTION ESP - BAD MU ') 
  105 ESP=BETAMU(IM) 
      RETURN 
      END 
C*********************************************************************
C************************* END OF ESP ********************************
C*********************************************************************
      FUNCTION EPRAY(TAU) 
C
C INPUT PARAMETER IS TAU AT LINE CENTER --- IE TAU/(R2PI*SIGNU) 
C  THE SUMMATION IS BASED ON AN EXACT INTEGRATION BUT IS NOT 
C   PRACTICAL FOR LARGE TAU. THE FORMULAS FOR LARGE TAU 
C   ARE FITS TO THE ANSWERS CALCULATED BY THE DCADRE ROUTINE 
C   AND ARE GENERALLY FAITHFUL TO 0.1 PERCENT. 
C
      INTEGER K
      REAL TAU,SUM,TERM,EPRAY
C
      IF(TAU.GT.6.0) GO TO 205 
      SUM=1.0-TAU/1.41421 
      TERM=-TAU/1.41421 
      DO 100 K=2,30 
         TERM=-TERM*TAU/FLOAT(K)*(FLOAT(K)/FLOAT(K+1))**0.5 
         SUM=SUM+TERM 
         IF(ABS(TERM).LT.1.0E-7) GO TO 105 
  100 CONTINUE 
  105 EPRAY=SUM 
      RETURN 
  205 CONTINUE 
      IF(TAU.GT.10.0) GO TO 215 
      EPRAY=0.22124/(TAU-0.96302*TAU**0.6) 
      RETURN 
  215 CONTINUE 
      IF(TAU.GT.25.0) GO TO 225 
      EPRAY=0.200165/(TAU-0.88216*TAU**0.7) 
      RETURN 
  225 CONTINUE 
      EPRAY=0.15397/(TAU-0.81841*TAU**0.84) 
      RETURN 
      END 
C*******************************************************************
C****************************** END OF EPRAY ***********************
C*******************************************************************
      SUBROUTINE TROUT(MOLCUL)
C
C   CALCULATES THE RADIAL OUTGOING EMISSION FROM THE SURFACE
C   FOR EACH OF THE MODEL TRANSITIONS AND EXPRESSES THEM IN
C   TR.
C 
      INTEGER I,J,ISUR
      REAL TEST,T0,T1,DTU,TAUTOT,RSUR,SSUR,SVAL,TBB,TR
      REAL TSTP,SUM,dtau,sout,gauss,epsln
      CHARACTER MOLCUL*10,LINE*80
      logical more
      EXTERNAL DTAU,SOUT,GAUSS
      data more / .true. /
C
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      INCLUDE 'mic_radsup.h'
      INCLUDE 'mic_tau.h'
C
      call LogWrite(' ',more)
      call LogWrite(' ',more)
      WRITE(line,2000) MOLCUL 
      call LogWrite(line,more)
      call LogWrite(' ',more)
      WRITE(line,2005)
      call LogWrite(line,more)
      RMU    = 0.0
      RADIUS = 0.0
      DO 100 JUP=2,NLEV 
      DO 100 JLO=1,NLEV 
         IF(A(JUP,JLO).EQ.0.0) GO TO 100 
         TSTP=RPT(NPTS)/(MAXT-1)
         SUM=0.0 
         AT(1)=0.0 
         ATAU(1)=0.0 
         DO 50 I=2,MAXT
            T0=TSTP*FLOAT(I-2) 
            T1=TSTP*FLOAT(I-1) 
            DTU=GAUSS(T0,T1,3,DTAU) 
            SUM=SUM+DTU 
            AT(I)=T1 
   50    ATAU(I)=SUM 
         TAUTOT=2.0*SUM 
         epsln = 1.0e-6
         CALL SPLNC(MAXT,AT,ATAU,epsln,AA) 
         DO 70 I=1,MAXT-1
            ISUR=I 
            TEST=SUM-ATAU(I) 
            IF(TEST.LE.1.5) GO TO 75 
   70    CONTINUE 
   75    RSUR=AT(ISUR) 
         IF(ISUR.EQ.MAXT-1) 
     1		RSUR=AT(ISUR)+1.5*(AT(MAXT)-AT(MAXT-1))/TEST 
         SSUR = GAUSS(RSUR,RPT(NPTS),15,SOUT) 
         SVAL = GAUSS(-RPT(NPTS),RSUR,15,SOUT) 
         TBB  = (EXP(-TAUTOT)-1.0)*BBK(JUP,JLO)+SVAL+SSUR 
         I    = JUP - 1 
         J    = JLO - 1 
         TR   = 0.5E-18*((AC/FREQ(JUP,JLO))**2)*TBB/AK 
         WRITE(line,2010) I,J,TAUTOT,TR,FREQ(JUP,JLO) 
         call LogWrite(line,more)
  100 CONTINUE 
 2000 FORMAT(12X,A5,1X,35HLINE RADIATION FOR CENTER OF SPHERE)
 2005 FORMAT(5X,2HJU,2X,2HJL,8X,3HTAU,10X,8HRAD TEMP,9X,4HFREQ) 
 2010 FORMAT(5X,I2,2X,I2,5X,1PE10.3,5X,E10.3,5X,0PF9.3) 
      RETURN 
      END 
C******************************************************************
C************************* END OF TROUT ***************************
C******************************************************************
      FUNCTION SOUT(XLEN) 
C
C   CALCULATES THE EMISSION FROM A LAYER AS SEEN FROM THE
C   SURFACE.
C
C INPUT: XLEN   - 
C        SOUR   - SOURCE FUNCTION FOR LAYER
C
C
      INCLUDE 'rad_size.h'
      INCLUDE 'mic_radsup.h'
      INCLUDE 'mic_tau.h'
C
      REAL XLEN,AXL,SDT,SOUT,TAU,SOUDTAU
      REAL XT(5),YT(5)
      EXTERNAL SOUDTAU
C
      AXL=ABS(XLEN) 
      XT(1)=AXL 
      CALL SPLN(MAXT,1,AT,ATAU,XT,YT,AA) 
      IF(XLEN.LE.0.0) TAU=ATAU(MAXT)+YT(1) 
      IF(XLEN.GT.0.0) TAU=ATAU(MAXT)-YT(1) 
      SDT=SOUDTAU(AXL) 
      SOUT=SDT*EXP(-TAU)
      RETURN 
      END 
C****************************************************************
C***************************** END OF SOUT **********************
C****************************************************************
      SUBROUTINE writeout(popname,molcul,pout) 
C
C  WRITES PARAMETERS WHICH FULLY DESCRIBE THE RESULTS OF THE MODEL
C  TO a MIRIAD file. THIS FILE COULD BE READ BY ANOTHER PROGRAM
C  TO CALCULATE THE ACTUAL EMISSION SEEN UNDER SPECIFIC OBSERVING
C  CONDITIONS.
C
      INTEGER i,j,k,pout,nsize(3)
      REAL crval(3),crdel(3)
      CHARACTER molcul*10,ctyp(3)*8,popname*50
C
      INCLUDE 'rad_size.h'
      INCLUDE 'radiat.h'
      INCLUDE 'mic_radtr.h'
      real row(MAXMOD)
C
      nsize(1) = npts
      nsize(2) = nlev
      nsize(3) = 1
      ctyp(1)  = 'radius'
      ctyp(2)  = 'levels'
      ctyp(3)  = 'dummy'
      crval(1) = 1.0
      crval(2) = 1.0
      crval(3) = 1.0
      crdel(1) = 1.0
      crdel(2) = 1.0
      crdel(3) = 1.0
      call xyopen(pout,popname,'new',3,nsize)
      call wrhead(pout,molcul,nsize,ctyp,crval,crdel)
C
C   write out populations
C
      do k=1,nsize(3)
         call xysetpl(pout,1,k)
         do j=1,nsize(2)
            do i=1,nsize(1)
               row(i) = popn(j,i)
            enddo
            call xywrite(pout,j,row)
         enddo
      enddo
      RETURN 
      END 
C**************************************************************
C************************ END OF WRITEOUT *********************
C**************************************************************
        subroutine wrhead(lout,molcul,nsize,ctyp,crval,crdel)
C
C  writes header for image cube
C
        integer lout,nsize(3)
        real crval(3),crdel(3)
        character molcul*10,ctyp(3)*8
C
        call wrhdi(lout,'naxis',3)
        call wrhdi(lout,'naxis1',nsize(1))
        call wrhdi(lout,'naxis2',nsize(2))
        call wrhdi(lout,'naxis3',nsize(3))
        call wrhda(lout,'ctype1',ctyp(1))
        call wrhda(lout,'ctype2',ctyp(2))
        call wrhda(lout,'ctype3',ctyp(3))
        call wrhdd(lout,'crpix1',1.0d0)
        call wrhdd(lout,'crpix2',1.0d0)
        call wrhdd(lout,'crpix3',1.0d0)
        call wrhdd(lout,'cdelt1',dble(crdel(1)))
        call wrhdd(lout,'cdelt2',dble(crdel(2)))
        call wrhdd(lout,'cdelt3',dble(crdel(3)))
        call wrhdd(lout,'crval1',dble(crval(1)))
        call wrhdd(lout,'crval2',dble(crval(2)))
        call wrhdd(lout,'crval3',dble(crval(3)))
        call wrhda(lout,'object',molcul)
        return
        end
C***************************************************************
C******************** END of WRHEAD ****************************
C***************************************************************
