	program radchi
c
c   Calculates the chi-square values for input observational line 
c   temperatures versus a microturbulent model.
c
c  History:
c    lgm 03jul92   Beginning writing of program from scratch.
c    lgm 01jun94   added logclose to end of main and logical keyprsnt
c    lgm 06aug97   updated log writes to be compatable with next miriad subs
c    pjt 02jul02   ansi fixed for linux
c
c------ Inline doc (retrieved with doc to a .doc file) ----------------c
c
c= RADCHI - Calculates chi-square between data and a microturbulent model.
c& lgm
c: models
c+
c  Calculates the chi-square values for input observational line
c  temperatures against a microturbulent model. The models are
c  assumed to be generated by the program RADSPH.
c@ data
c   Name of file containing observational data in the following
c   format: radial-position freq line-temp temp-error
c           radial-position freq line-temp temp-error
c           ............... .... ......... ..........
c   Maximum of 200 data points.
c   No default.
c@ model
c   Name of file containing model for chi-square calculation.
c   This model grid should have been generated by RADSPH.
c   No default.
c@ log
c   Log file name. Default is to write to standard output.
c
c----------------------------------------------------------------
c
	INCLUDE 'rad_size.h'
c
	integer maxdata,maxnax,maxarr
	parameter (MAXDATA=200)
	parameter (MAXNAX=3)
	parameter (MAXARR=256)
	character datafile*50,logfile*50
	character version*25,infiles(10)*50
	character line*100,words(20)*20,molcul*10
	character ctyp(4)*8
	real crpix(4),crval(4),crdel(4)
	real datfreq(MAXDATA),data(MAXDATA)
        real daterr(MAXDATA),pos(MAXDATA)
	real tranfr(MAXTRAN),modval(MAXDATA),chisqr
	real newval,cube(MAXARR,MAXARR,MAXTRAN)
	integer i,j,in,naxis,nwords,iplane,nsize(4)
	integer idat,maxplane,ypix,ifil
	integer nfiles,pixpos(MAXDATA),which(MAXDATA)
	logical eof,more,keyprsnt
c
	data more /.true./
	DATA version / 'Version Jul 2, 2002' /
c
	call output('RADCHI:'//version)
c
c  get user inputs
c
	call keyini
	call keya('data',datafile,'  ')
	if(datafile(1:2) .eq. '  ') 
     1			call bug('f','DATA must be input')
	nfiles = 0
	do while (keyprsnt('model'))
	   nfiles = nfiles + 1
	   call keya('model',infiles(nfiles),'  ')
	enddo
	if(infiles(1)(1:2) .eq. '  ')
     1			call bug('f','MODEL must be input')
	call keya('log',logfile,' ')
	call LogOpen(logfile,' ')
	call keyfin
c
c  read in data file now
c
        open(unit=22,file=datafile,status='old')
        do i=1,MAXDATA
           read(22,*,end=105) pos(i),datfreq(i),data(i),
     1                  daterr(i),which(i)
           ndata = i
        enddo
  105   continue
c
c   Loop over input files
c
	do ifil=1,nfiles
c
c  open model file and get data
c
	   eof = .false.
	   nsize(4) = 1
	   naxis    = 3
	   call xyopen(in,infiles(ifil),'old',naxis,nsize)
	   call rdhead(in,naxis,molcul,ctyp,crval,crdel,crpix)
	   call rdgrid(in,naxis,nsize,cube)
c
c  read freq information from history file
c
	   maxplane = 0
	   call hisopen(in,'read')
	   do while (.not. eof) 
	      call hisread(in,line,eof)
	      call sopchr(line,words,nwords)
	      if(words(1)(1:6) .eq. 'RADSPH') then
	         if(words(2)(1:6) .eq. 'PLANE ') then
	  	    read(words(4),*) iplane
		    read(words(6),*) freq
		    tranfr(iplane) = freq
		    if(iplane .gt. maxplane) maxplane = iplane
	         endif
	      endif
	   enddo
	   call hisclose(in)
c
c  translate position input to pixel locations
c
           do i=1,ndata
              j = pos(i)/(206265.*abs(crdel(1)))
              pixpos(i) = crpix(1) - j
              if(pixpos(i) .lt. 1)
     1           call bug('f','Pixel out of range')
           enddo
c
c  pick-up model values
c  
	   ypix = crpix(2)
	   do idat=1,ndata
	      do i=1,maxplane
	         if(abs(datfreq(idat)-tranfr(i)) .lt. 0.05 .and.
     1		     ifil .eq. which(idat)) 
     1		       modval(idat) = cube(pixpos(idat),ypix,i)
	      enddo
	   enddo
	   call xyclose(in)
	enddo
c
c  calculate chi-square
	chisqr = 0.0
	do i=1,ndata
	   newval = (data(i)-modval(i))*(data(i)-modval(i))/
     1			(daterr(i)*daterr(i))
	   chisqr = chisqr + newval
	enddo
c
c   write out results
c
	call LogWrite('   Results of Chi-square fit:',more)
	do i=1,nfiles
	   write(line,'('' Using Model file: '',a)') infiles(i)
	   call LogWrite(line,more)
	enddo
	call LogWrite('  ',more)
	do i=1,ndata
	   write(line,'(''  Position: '',f5.1,'' Freq: '',f7.2,
     0        '' Obs. Tr: '',f6.2,
     1	      '' Model Tr: '',f6.2,'' sigma: '',f6.3)')
     2	       pos(i),datfreq(i),data(i),modval(i),daterr(i)
	   call Logwrite(line,more)
	enddo
	write(line,'(''   Chi-sqr: '',f8.2)') chisqr
	call Logwrite(line,more)

        call logclose
	stop
	end

	subroutine rdhead(in,naxis,molcul,ctyp,crval,crdel,crpix)
c
	character*(*) ctyp(4), molcul
	real crval(4),crdel(4),crpix(4)
	double precision dummy
c
	call rdhda(in,'object',molcul,' ')
	call rdhda(in,'ctype1',ctyp(1),' ')
        call rdhda(in,'ctype2',ctyp(2),' ')
        call rdhda(in,'ctype3',ctyp(3),' ')
	call rdhdd(in,'crpix1',dummy,0.0d0)
	crpix(1) = dummy
        call rdhdd(in,'crpix2',dummy,0.0d0)
        crpix(2) = dummy
        call rdhdd(in,'crpix3',dummy,0.0d0)
        crpix(3) = dummy
	call rdhdd(in,'cdelt1',dummy,0.0d0)
	crdel(1) = dummy
        call rdhdd(in,'cdelt2',dummy,0.0d0)
        crdel(2) = dummy
        call rdhdd(in,'cdelt3',dummy,0.0d0)
        crdel(3) = dummy
	call rdhdd(in,'crval1',dummy,0.0d0)
	crval(1) = dummy
        call rdhdd(in,'crval2',dummy,0.0d0)
        crval(2) = dummy
        call rdhdd(in,'crval3',dummy,0.0d0)
        crval(3) = dummy
	if(naxis .eq. 4) then
           call rdhda(in,'ctype4',ctyp(4),' ')
           call rdhdd(in,'crpix4',dummy,0.0d0)
           crpix(4) = dummy
           call rdhdd(in,'cdelt4',dummy,0.0d0)
           crdel(4) = dummy
           call rdhdd(in,'crval4',dummy,0.0d0)
           crval(4) = dummy
	endif
c
	return
	end
	
	subroutine rdgrid(in,naxis,nsize,cube)
c
	INCLUDE 'rad_size.h'
	integer maxarr
	parameter (MAXARR=256)
c
c

	integer in,naxis,k,j,i,nsize(3)
	real row(MAXARR),cube(MAXARR,MAXARR,MAXTRAN)
c
	do k=1,nsize(3)
	   call xysetpl(in,1,k)
	   do j=1,nsize(2)
	      call xyread(in,j,row)
	      do i=1,nsize(1)
	         cube(i,j,k) = row(i)
	      enddo
	   enddo
	enddo
	return
	end
